import json
import logging
import sys
from argparse import Namespace
from uuid import UUID

import pytest

from arkindex.exceptions import ErrorResponse
from arkindex_worker.cache import (
    SQL_VERSION,
    CachedElement,
    create_version_table,
    init_cache_db,
)
from arkindex_worker.models import Element
from arkindex_worker.worker import ActivityState, ElementsWorker
from arkindex_worker.worker.dataset import DatasetState
from arkindex_worker.worker.process import ProcessMode
from tests import PROCESS_ID

from . import BASE_API_CALLS


def test_database_arg(mocker, mock_elements_worker, tmp_path):
    database_path = tmp_path / "my_database.sqlite"
    init_cache_db(database_path)
    create_version_table()

    mocker.patch(
        "arkindex_worker.worker.base.argparse.ArgumentParser.parse_args",
        return_value=Namespace(
            element=["volumeid", "pageid"],
            verbose=False,
            elements_list=None,
            database=database_path,
            dev=False,
            set=[],
        ),
    )

    worker = ElementsWorker(support_cache=True)
    worker.configure()

    assert worker.use_cache is True
    assert worker.cache_path == database_path


def test_database_arg_cache_missing_version_table(
    mocker, mock_elements_worker, tmp_path
):
    database_path = tmp_path / "my_database.sqlite"
    database_path.touch()

    mocker.patch(
        "arkindex_worker.worker.base.argparse.ArgumentParser.parse_args",
        return_value=Namespace(
            element=["volumeid", "pageid"],
            verbose=False,
            elements_list=None,
            database=database_path,
            dev=False,
            set=[],
        ),
    )

    worker = ElementsWorker(support_cache=True)
    with pytest.raises(
        AssertionError,
        match=f"The SQLite database {database_path} does not have the correct cache version, it should be {SQL_VERSION}",
    ):
        worker.configure()


def test_readonly(responses, mock_elements_worker):
    """Test readonly worker does not trigger any API calls"""

    # Setup the worker as read-only
    mock_elements_worker.worker_run_id = None
    assert mock_elements_worker.is_read_only is True

    out = mock_elements_worker.update_activity("1234-deadbeef", ActivityState.Processed)

    # update_activity returns False in very specific cases
    assert out is True
    assert len(responses.calls) == len(BASE_API_CALLS)
    assert [
        (call.request.method, call.request.url) for call in responses.calls
    ] == BASE_API_CALLS


def test_get_elements_elements_list_arg_wrong_type(
    monkeypatch, tmp_path, mock_elements_worker
):
    elements_path = tmp_path / "elements.json"
    elements_path.write_text("{}")

    monkeypatch.setenv("TASK_ELEMENTS", str(elements_path))
    worker = ElementsWorker()
    worker.configure()

    with pytest.raises(AssertionError, match="Elements list must be a list"):
        worker.get_elements()


def test_get_elements_elements_list_arg_empty_list(
    monkeypatch, tmp_path, mock_elements_worker
):
    elements_path = tmp_path / "elements.json"
    elements_path.write_text("[]")

    monkeypatch.setenv("TASK_ELEMENTS", str(elements_path))
    worker = ElementsWorker()
    worker.configure()

    with pytest.raises(AssertionError, match="No elements in elements list"):
        worker.get_elements()


def test_get_elements_elements_list_arg_missing_id(
    monkeypatch, tmp_path, mock_elements_worker
):
    elements_path = tmp_path / "elements.json"
    elements_path.write_text(json.dumps([{"type": "volume"}]))

    monkeypatch.setenv("TASK_ELEMENTS", str(elements_path))
    worker = ElementsWorker()
    worker.configure()

    elt_list = worker.get_elements()

    assert elt_list == []


def test_get_elements_elements_list_arg_not_uuid(
    monkeypatch, tmp_path, mock_elements_worker
):
    elements_path = tmp_path / "elements.json"
    elements_path.write_text(
        json.dumps(
            [
                {"id": "volumeid", "type": "volume"},
                {"id": "pageid", "type": "page"},
                {"id": "actid", "type": "act"},
                {"id": "surfaceid", "type": "surface"},
            ]
        )
    )

    monkeypatch.setenv("TASK_ELEMENTS", str(elements_path))
    worker = ElementsWorker()
    worker.configure()

    with pytest.raises(
        Exception,
        match="These element IDs are invalid: volumeid, pageid, actid, surfaceid",
    ):
        worker.get_elements()


def test_get_elements_elements_list_arg(monkeypatch, tmp_path, mock_elements_worker):
    elements_path = tmp_path / "elements.json"
    elements_path.write_text(
        json.dumps(
            [
                {"id": "11111111-1111-1111-1111-111111111111", "type": "volume"},
                {"id": "22222222-2222-2222-2222-222222222222", "type": "page"},
                {"id": "33333333-3333-3333-3333-333333333333", "type": "act"},
            ]
        )
    )

    monkeypatch.setenv("TASK_ELEMENTS", str(elements_path))
    worker = ElementsWorker()
    worker.configure()

    elt_list = worker.get_elements()

    assert elt_list == [
        "11111111-1111-1111-1111-111111111111",
        "22222222-2222-2222-2222-222222222222",
        "33333333-3333-3333-3333-333333333333",
    ]


def test_get_elements_element_arg_not_uuid(mocker, mock_elements_worker):
    mocker.patch(
        "arkindex_worker.worker.base.argparse.ArgumentParser.parse_args",
        return_value=Namespace(
            element=["volumeid", "pageid"],
            config={},
            verbose=False,
            elements_list=None,
            database=None,
            dev=True,
            set=[],
        ),
    )

    worker = ElementsWorker()
    worker.configure()

    with pytest.raises(
        Exception, match="These element IDs are invalid: volumeid, pageid"
    ):
        worker.get_elements()


def test_get_elements_element_arg(mocker, mock_elements_worker):
    mocker.patch(
        "arkindex_worker.worker.base.argparse.ArgumentParser.parse_args",
        return_value=Namespace(
            element=[
                "11111111-1111-1111-1111-111111111111",
                "22222222-2222-2222-2222-222222222222",
            ],
            config={},
            verbose=False,
            elements_list=None,
            database=None,
            dev=True,
            set=[],
        ),
    )

    worker = ElementsWorker()
    worker.configure()

    elt_list = worker.get_elements()

    assert elt_list == [
        "11111111-1111-1111-1111-111111111111",
        "22222222-2222-2222-2222-222222222222",
    ]


def test_get_elements_dataset_set_arg(responses, mocker, mock_elements_worker):
    mocker.patch(
        "arkindex_worker.worker.base.argparse.ArgumentParser.parse_args",
        return_value=Namespace(
            element=[],
            config={},
            verbose=False,
            elements_list=None,
            database=None,
            dev=True,
            set=[(UUID("11111111-1111-1111-1111-111111111111"), "train")],
        ),
    )

    # Mock RetrieveDataset call
    responses.add(
        responses.GET,
        "http://testserver/api/v1/datasets/11111111-1111-1111-1111-111111111111/",
        status=200,
        json={
            "id": "11111111-1111-1111-1111-111111111111",
            "name": "My dataset",
            "description": "A dataset about cats.",
            "sets": ["train", "dev", "test"],
            "state": DatasetState.Complete.value,
        },
        content_type="application/json",
    )

    # Mock ListSetElements call
    element = {
        "id": "22222222-2222-2222-2222-222222222222",
        "type": "page",
        "name": "1",
        "corpus": {
            "id": "11111111-1111-1111-1111-111111111111",
        },
        "thumbnail_url": "http://example.com",
        "zone": {
            "id": "497f6eca-6276-4993-bfeb-53cbbbba6f08",
            "polygon": [[0, 0], [0, 0], [0, 0]],
            "image": {
                "id": "497f6eca-6276-4993-bfeb-53cbbbba6f08",
                "path": "string",
                "width": 0,
                "height": 0,
                "url": "http://example.com",
                "s3_url": "string",
                "status": "checked",
                "server": {
                    "display_name": "string",
                    "url": "http://example.com",
                    "max_width": 2147483647,
                    "max_height": 2147483647,
                },
            },
            "url": "http://example.com",
        },
        "rotation_angle": 0,
        "mirrored": False,
        "created": "2019-08-24T14:15:22Z",
        "classes": [
            {
                "id": "497f6eca-6276-4993-bfeb-53cbbbba6f08",
                "ml_class": {
                    "id": "497f6eca-6276-4993-bfeb-53cbbbba6f08",
                    "name": "string",
                },
                "state": "pending",
                "confidence": 0,
                "high_confidence": True,
                "worker_run": {
                    "id": "497f6eca-6276-4993-bfeb-53cbbbba6f08",
                    "summary": "string",
                },
            }
        ],
        "metadata": [
            {
                "id": "497f6eca-6276-4993-bfeb-53cbbbba6f08",
                "type": "text",
                "name": "string",
                "value": "string",
                "dates": [{"type": "exact", "year": 0, "month": 1, "day": 1}],
            }
        ],
        "transcriptions": [
            {
                "id": "497f6eca-6276-4993-bfeb-53cbbbba6f08",
                "text": "string",
                "confidence": 0,
                "orientation": "horizontal-lr",
                "worker_run": {
                    "id": "497f6eca-6276-4993-bfeb-53cbbbba6f08",
                    "summary": "string",
                },
            }
        ],
        "has_children": True,
        "worker_run": {
            "id": "497f6eca-6276-4993-bfeb-53cbbbba6f08",
            "summary": "string",
        },
        "confidence": 1,
    }
    responses.add(
        responses.GET,
        "http://testserver/api/v1/datasets/11111111-1111-1111-1111-111111111111/elements/?set=train&with_count=true",
        status=200,
        json={
            "next": None,
            "previous": None,
            "results": [
                {
                    "set": "train",
                    "element": element,
                }
            ],
            "count": 1,
        },
        content_type="application/json",
    )

    worker = ElementsWorker()
    worker.configure()

    elt_list = worker.get_elements()

    assert elt_list == [
        Element(**element),
    ]


def test_get_elements_dataset_set_api(responses, mocker, mock_elements_worker):
    # Mock ListProcessSets call
    responses.add(
        responses.GET,
        "http://testserver/api/v1/process/aaaaaaaa-bbbb-cccc-dddd-eeeeeeeeffff/sets/",
        status=200,
        json={
            "next": None,
            "previous": None,
            "results": [
                {
                    "id": "33333333-3333-3333-3333-333333333333",
                    "dataset": {"id": "11111111-1111-1111-1111-111111111111"},
                    "set_name": "train",
                }
            ],
            "count": 1,
        },
        content_type="application/json",
    )

    # Mock ListSetElements call
    element = {
        "id": "22222222-2222-2222-2222-222222222222",
        "type": "page",
        "name": "1",
        "corpus": {
            "id": "11111111-1111-1111-1111-111111111111",
        },
        "thumbnail_url": "http://example.com",
        "zone": {
            "id": "497f6eca-6276-4993-bfeb-53cbbbba6f08",
            "polygon": [[0, 0], [0, 0], [0, 0]],
            "image": {
                "id": "497f6eca-6276-4993-bfeb-53cbbbba6f08",
                "path": "string",
                "width": 0,
                "height": 0,
                "url": "http://example.com",
                "s3_url": "string",
                "status": "checked",
                "server": {
                    "display_name": "string",
                    "url": "http://example.com",
                    "max_width": 2147483647,
                    "max_height": 2147483647,
                },
            },
            "url": "http://example.com",
        },
        "rotation_angle": 0,
        "mirrored": False,
        "created": "2019-08-24T14:15:22Z",
        "classes": [
            {
                "id": "497f6eca-6276-4993-bfeb-53cbbbba6f08",
                "ml_class": {
                    "id": "497f6eca-6276-4993-bfeb-53cbbbba6f08",
                    "name": "string",
                },
                "state": "pending",
                "confidence": 0,
                "high_confidence": True,
                "worker_run": {
                    "id": "497f6eca-6276-4993-bfeb-53cbbbba6f08",
                    "summary": "string",
                },
            }
        ],
        "metadata": [
            {
                "id": "497f6eca-6276-4993-bfeb-53cbbbba6f08",
                "type": "text",
                "name": "string",
                "value": "string",
                "dates": [{"type": "exact", "year": 0, "month": 1, "day": 1}],
            }
        ],
        "transcriptions": [
            {
                "id": "497f6eca-6276-4993-bfeb-53cbbbba6f08",
                "text": "string",
                "confidence": 0,
                "orientation": "horizontal-lr",
                "worker_run": {
                    "id": "497f6eca-6276-4993-bfeb-53cbbbba6f08",
                    "summary": "string",
                },
            }
        ],
        "has_children": True,
        "worker_run": {
            "id": "497f6eca-6276-4993-bfeb-53cbbbba6f08",
            "summary": "string",
        },
        "confidence": 1,
    }
    responses.add(
        responses.GET,
        "http://testserver/api/v1/datasets/11111111-1111-1111-1111-111111111111/elements/?set=train&with_count=true",
        status=200,
        json={
            "next": None,
            "previous": None,
            "results": [
                {
                    "set": "train",
                    "element": element,
                }
            ],
            "count": 1,
        },
        content_type="application/json",
    )

    # Update ProcessMode to Dataset
    mock_elements_worker.process_information["mode"] = ProcessMode.Dataset

    elt_list = mock_elements_worker.get_elements()

    assert elt_list == [
        Element(**element),
    ]


def test_get_elements_both_args_error(mocker, mock_elements_worker, tmp_path):
    elements_path = tmp_path / "elements.json"
    elements_path.write_text(
        json.dumps(
            [
                {"id": "volumeid", "type": "volume"},
                {"id": "pageid", "type": "page"},
                {"id": "actid", "type": "act"},
                {"id": "surfaceid", "type": "surface"},
            ]
        )
    )
    mocker.patch(
        "arkindex_worker.worker.base.argparse.ArgumentParser.parse_args",
        return_value=Namespace(
            element=["anotherid", "againanotherid"],
            verbose=False,
            elements_list=elements_path.open(),
            database=None,
            dev=False,
            set=[],
        ),
    )

    worker = ElementsWorker()
    worker.configure()

    with pytest.raises(
        AssertionError, match="elements-list and element CLI args shouldn't be both set"
    ):
        worker.get_elements()


def test_get_elements_export_process(mock_elements_worker, responses):
    responses.add(
        responses.GET,
        f"http://testserver/api/v1/process/{PROCESS_ID}/elements/?page_size=500&with_count=true&with_image=False",
        status=200,
        json={
            "count": 2,
            "next": None,
            "results": [
                {
                    "id": "aaaaaaaa-aaaa-aaaa-aaaa-aaaaaaaaaaaa",
                    "type_id": "baaaaaaa-aaaa-aaaa-aaaa-aaaaaaaaaaaa",
                    "name": "element 1",
                    "confidence": 1,
                    "image_id": None,
                    "image_width": None,
                    "image_height": None,
                    "image_url": None,
                    "polygon": None,
                    "rotation_angle": 0,
                    "mirrored": False,
                },
                {
                    "id": "aaaaaaaa-aaaa-aaaa-aaaa-aaaaaaaaaaa0",
                    "type_id": "baaaaaaa-aaaa-aaaa-aaaa-aaaaaaaaaaaa",
                    "name": "element 2",
                    "confidence": 1,
                    "image_id": None,
                    "image_width": None,
                    "image_height": None,
                    "image_url": None,
                    "polygon": None,
                    "rotation_angle": 0,
                    "mirrored": False,
                },
            ],
        },
    )
    mock_elements_worker.process_information["mode"] = "export"
    assert set(mock_elements_worker.get_elements()) == {
        "aaaaaaaa-aaaa-aaaa-aaaa-aaaaaaaaaaaa",
        "aaaaaaaa-aaaa-aaaa-aaaa-aaaaaaaaaaa0",
    }


@pytest.mark.usefixtures("_mock_worker_run_api")
def test_activities_disabled(responses, monkeypatch):
    """Test worker process elements without updating activities when they are disabled for the process"""
    monkeypatch.setattr(sys, "argv", ["worker"])
    worker = ElementsWorker()
    worker.configure()
    assert not worker.is_read_only

    assert len(responses.calls) == len(BASE_API_CALLS)
    assert [
        (call.request.method, call.request.url) for call in responses.calls
    ] == BASE_API_CALLS


def test_activities_dev_mode(mocker):
    """
    Worker activities are not stored in dev mode
    """
    worker = ElementsWorker()
    mocker.patch.object(sys, "argv", ["worker", "--dev"])
    worker.configure()

    assert worker.args.dev is True
    assert worker.process_information is None
    assert worker.is_read_only is True
    assert worker.store_activity is False


@pytest.mark.usefixtures("_mock_activity_calls")
@pytest.mark.parametrize(
    ("process_exception", "final_state"),
    [
        # Successful process_element
        (None, "processed"),
        # Failures in process_element
        (
            ErrorResponse(title="bad gateway", status_code=502, content="Bad gateway"),
            "error",
        ),
        (ValueError("Something bad"), "error"),
        (Exception("Any error"), "error"),
    ],
)
def test_run(
    monkeypatch,
    mock_elements_worker_with_list,
    responses,
    process_exception,
    final_state,
):
    """Check the normal runtime sends 2 API calls to update activity"""
    # Disable second configure call from run()
    monkeypatch.setattr(mock_elements_worker_with_list, "configure", lambda: None)
    assert mock_elements_worker_with_list.is_read_only is False
    # Mock exception in process_element
    if process_exception:

        def _err():
            raise process_exception

        monkeypatch.setattr(mock_elements_worker_with_list, "process_element", _err)

        # The worker stops because all elements failed !
        with pytest.raises(SystemExit):
            mock_elements_worker_with_list.run()
    else:
        # Simply run the process
        mock_elements_worker_with_list.run()

    assert len(responses.calls) == len(BASE_API_CALLS) + 3
    assert [
        (call.request.method, call.request.url) for call in responses.calls
    ] == BASE_API_CALLS + [
        ("GET", "http://testserver/api/v1/element/1234-deadbeef/"),
        (
            "PUT",
            "http://testserver/api/v1/workers/versions/56785678-5678-5678-5678-567856785678/activity/",
        ),
        (
            "PUT",
            "http://testserver/api/v1/workers/versions/56785678-5678-5678-5678-567856785678/activity/",
        ),
    ]

    # Check the requests sent by worker
    assert json.loads(responses.calls[-2].request.body) == {
        "element_id": "1234-deadbeef",
        "process_id": "aaaaaaaa-bbbb-cccc-dddd-eeeeeeeeffff",
        "state": "started",
    }
    assert json.loads(responses.calls[-1].request.body) == {
        "element_id": "1234-deadbeef",
        "process_id": "aaaaaaaa-bbbb-cccc-dddd-eeeeeeeeffff",
        "state": final_state,
    }


@pytest.mark.usefixtures("_mock_cached_elements", "_mock_activity_calls")
def test_run_cache(monkeypatch, mocker, mock_elements_worker_with_cache):
    # Disable second configure call from run()
    monkeypatch.setattr(mock_elements_worker_with_cache, "configure", lambda: None)

    # Make all the cached elements from the fixture initial elements
    CachedElement.update(initial=True).execute()

    mock_elements_worker_with_cache.process_element = mocker.MagicMock()
    mock_elements_worker_with_cache.run()

    assert mock_elements_worker_with_cache.process_element.call_args_list == [
        # Called once for each cached element
        mocker.call(elt)
        for elt in CachedElement.select()
    ]


def test_run_consuming_worker_activities(
    monkeypatch,
    mock_elements_worker_consume_wa,
    responses,
    caplog,
):
    """Check the consuming worker activities runtime uses StartWorkerActivity + UpdateWorkerActivity"""
    # Disable second configure call from run()
    monkeypatch.setattr(mock_elements_worker_consume_wa, "configure", lambda: None)

    assert mock_elements_worker_consume_wa.is_read_only is False

    # Provide 2 worker activities to run and the corresponding update call
    # and 2 element details response
    for i, elt_id in enumerate(("page_1", "page_2"), 1):
        responses.add(
            responses.POST,
            "http://testserver/api/v1/process/start-activity/",
            status=200,
            json={
                "id": elt_id,
                "type_id": "page-aaaa-aaaa-aaaa-aaaaaaaaaaaa",  # Element type provided by mock corpus
                "name": f"Page n°{i}",
            },
        )
        responses.add(
            responses.PUT,
            "http://testserver/api/v1/workers/versions/56785678-5678-5678-5678-567856785678/activity/",
            status=200,
        )
        responses.add(
            responses.GET,
            f"http://testserver/api/v1/element/{elt_id}/",
            status=200,
            json={
                "id": elt_id,
                "type": "page",
                "name": f"Page n°{i}",
            },
        )

    # Then a 404 to stop iterating
    responses.add(
        responses.POST,
        "http://testserver/api/v1/process/start-activity/",
        status=404,
    )

    # Simply run the process
    mock_elements_worker_consume_wa.run()

    # We call twice configure in the conftest
    assert len(responses.calls) == len(BASE_API_CALLS) * 2 + 7
    assert [
        (call.request.method, call.request.url) for call in responses.calls
    ] == BASE_API_CALLS * 2 + [
        (
            "POST",
            "http://testserver/api/v1/process/start-activity/",
        ),
        (
            "GET",
            "http://testserver/api/v1/element/page_1/",
        ),
        (
            "PUT",
            "http://testserver/api/v1/workers/versions/56785678-5678-5678-5678-567856785678/activity/",
        ),
        (
            "POST",
            "http://testserver/api/v1/process/start-activity/",
        ),
        (
            "GET",
            "http://testserver/api/v1/element/page_2/",
        ),
        (
            "PUT",
            "http://testserver/api/v1/workers/versions/56785678-5678-5678-5678-567856785678/activity/",
        ),
        (
            "POST",
            "http://testserver/api/v1/process/start-activity/",
        ),
    ]

    assert [(record.levelno, record.message) for record in caplog.records] == [
        (
            logging.INFO,
            "Using StartWorkerActivity instead of reading init_elements JSON file",
        ),
        (
            logging.INFO,
            "Processing page Page n°1 (page_1) (n°1)",
        ),
        (
            logging.INFO,
            "Processing page Page n°2 (page_2) (n°2)",
        ),
        (
            logging.INFO,
            "Ran on 2 elements: 2 completed, 0 failed",
        ),
    ]


def test_start_activity_conflict(
    monkeypatch, responses, mock_elements_worker_with_list, caplog
):
    # Disable second configure call from run()
    monkeypatch.setattr(mock_elements_worker_with_list, "configure", lambda: None)

    # Mock a "normal" conflict during in activity update, which returns the Exception
    responses.add(
        responses.PUT,
        "http://testserver/api/v1/workers/versions/56785678-5678-5678-5678-567856785678/activity/",
        body=ErrorResponse(
            title="conflict",
            status_code=409,
            content="Either this activity does not exists or this state is not allowed.",
        ),
    )

    mock_elements_worker_with_list.run()

    assert len(responses.calls) == len(BASE_API_CALLS) + 2
    assert [
        (call.request.method, call.request.url) for call in responses.calls
    ] == BASE_API_CALLS + [
        ("GET", "http://testserver/api/v1/element/1234-deadbeef/"),
        (
            "PUT",
            "http://testserver/api/v1/workers/versions/56785678-5678-5678-5678-567856785678/activity/",
        ),
    ]
    assert [(record.levelno, record.message) for record in caplog.records] == [
        (logging.INFO, "Processing page Test Page n°1 (1234-deadbeef) (1/1)"),
        (logging.INFO, "Skipping element 1234-deadbeef as it was already processed"),
        (logging.INFO, "Ran on 1 element: 1 completed, 0 failed"),
    ]


def test_start_activity_error(
    monkeypatch, responses, mock_elements_worker_with_list, caplog
):
    # Disable second configure call from run()
    monkeypatch.setattr(mock_elements_worker_with_list, "configure", lambda: None)

    # Mock a random error occurring during the activity update
    responses.add(
        responses.PUT,
        "http://testserver/api/v1/workers/versions/56785678-5678-5678-5678-567856785678/activity/",
        body=Exception("A wild Petilil appears!"),
    )

    with pytest.raises(SystemExit):
        mock_elements_worker_with_list.run()

    assert [
        (call.request.method, call.request.url) for call in responses.calls
    ] == BASE_API_CALLS + [
        ("GET", "http://testserver/api/v1/element/1234-deadbeef/"),
        (
            "PUT",
            "http://testserver/api/v1/workers/versions/56785678-5678-5678-5678-567856785678/activity/",
        ),
        # Activity is updated to the "error" state regardless of the Exception occurring during the call
        (
            "PUT",
            "http://testserver/api/v1/workers/versions/56785678-5678-5678-5678-567856785678/activity/",
        ),
    ]
    assert [(record.levelno, record.message) for record in caplog.records] == [
        (logging.INFO, "Processing page Test Page n°1 (1234-deadbeef) (1/1)"),
        (
            logging.WARNING,
            "Failed running worker on element 1234-deadbeef: Exception('A wild Petilil appears!')",
        ),
        (logging.ERROR, "Ran on 1 element: 0 completed, 1 failed"),
    ]


@pytest.mark.usefixtures("_mock_worker_run_api")
def test_update_activity(responses, mock_elements_worker):
    """Test an update call with feature enabled triggers an API call"""
    responses.add(
        responses.PUT,
        "http://testserver/api/v1/workers/versions/56785678-5678-5678-5678-567856785678/activity/",
        status=200,
        json={
            "element_id": "1234-deadbeef",
            "process_id": "aaaaaaaa-bbbb-cccc-dddd-eeeeeeeeffff",
            "state": "processed",
        },
    )

    out = mock_elements_worker.update_activity("1234-deadbeef", ActivityState.Processed)

    # Check the response received by worker
    assert out is True

    assert len(responses.calls) == len(BASE_API_CALLS) + 1
    assert [
        (call.request.method, call.request.url) for call in responses.calls
    ] == BASE_API_CALLS + [
        (
            "PUT",
            "http://testserver/api/v1/workers/versions/56785678-5678-5678-5678-567856785678/activity/",
        ),
    ]

    # Check the request sent by worker
    assert json.loads(responses.calls[-1].request.body) == {
        "element_id": "1234-deadbeef",
        "process_id": "aaaaaaaa-bbbb-cccc-dddd-eeeeeeeeffff",
        "state": "processed",
    }



"""
ElementsWorker methods for elements and element types.
"""

import os
from collections.abc import Iterable
from operator import attrgetter
from uuid import UUID
from warnings import warn

from peewee import IntegrityError

from arkindex.exceptions import ErrorResponse
from arkindex_worker import logger
from arkindex_worker.cache import CachedElement, CachedImage, unsupported_cache
from arkindex_worker.models import Element
from arkindex_worker.utils import (
    DEFAULT_BATCH_SIZE,
    batch_publication,
    make_batches,
    pluralize,
)


class MissingElementType(Exception):
    """
    Raised when the specified element type was not found in the corpus and
    the worker cannot create it.
    """


class ElementMixin:
    def add_arguments(self):
        """Define specific ``argparse`` arguments for the worker using this mixin"""
        self.parser.add_argument(
            "--elements-list",
            help="JSON elements list to use",
            type=open,
            default=os.environ.get("TASK_ELEMENTS"),
        )
        self.parser.add_argument(
            "--no-elements-list",
            help=(
                "Consume worker activities from Arkindex API instead of using a static elements list"
            ),
            dest="consume_worker_activities",
            action="store_true",
            default=os.environ.get("SKIP_TASK_ELEMENTS") is not None,
        )
        self.parser.add_argument(
            "--element",
            type=str,
            nargs="+",
            help="One or more Arkindex element ID",
        )
        super().add_arguments()

    @property
    def consume_worker_activities(self) -> bool:
        """
        Helper to detect if the worker rely on an elements.json or consume directly worker activities
        Uses the process information when available, fallback to CLI args
        """
        if self.process_information is not None:
            return self.process_information.get("skip_elements_json") is True

        return self.args.consume_worker_activities

    def list_corpus_types(self):
        """
        Loads available element types in corpus.
        """
        self.corpus_types = {
            element_type["slug"]: element_type
            for element_type in self.api_client.request(
                "RetrieveCorpus", id=self.corpus_id
            )["types"]
        }
        count = len(self.corpus_types)
        logger.info(
            f"Loaded {count} element {pluralize('type', count)} in corpus ({self.corpus_id})."
        )

    @unsupported_cache
    def create_element_type(
        self, slug: str, name: str, is_folder: bool = False
    ) -> None:
        """
        Create an element type on the given corpus.

        :param slug: Slug of the element type.
        :param name: Name of the element type.
        :param is_folder: Whether an element with this type can contain other elements or not.
        """
        assert slug and isinstance(slug, str), (
            "slug shouldn't be null and should be of type str"
        )
        assert name and isinstance(name, str), (
            "name shouldn't be null and should be of type str"
        )
        assert is_folder is not None and isinstance(is_folder, bool), (
            "is_folder shouldn't be null and should be of type bool"
        )

        try:
            element_type = self.api_client.request(
                "CreateElementType",
                body={
                    "slug": slug,
                    "display_name": name,
                    "folder": is_folder,
                    "corpus": self.corpus_id,
                },
            )
            self.corpus_types[slug] = element_type
            logger.info(f"Created a new element type with slug `{slug}`.")
        except ErrorResponse as e:
            # Only reload for 400 errors
            if e.status_code != 400:
                raise

            # Reload and make sure we have the element type now
            logger.warning(
                f"Unable to create the element type `{slug}`. Refreshing corpus element types cache."
            )
            self.list_corpus_types()
            assert slug in self.corpus_types, (
                f"Missing element type `{slug}` even after refreshing."
            )

    def check_required_types(
        self, type_slugs: list[str], create_missing: bool = False
    ) -> None:
        """
        Check that every element type needed is available in the corpus.
        Missing ones may be created automatically if needed.

        :param type_slugs: Element type slugs to search.
        :param create_missing: Whether the missing types should be created. Defaults to False.
        :raises MissingElementType: When an entity type is missing and cannot be created.
        """
        assert type_slugs and isinstance(type_slugs, list), (
            "type_slugs shouldn't be null and should be of type list"
        )

        for index, slug in enumerate(type_slugs):
            assert isinstance(slug, str), (
                f"Element type at index {index} in type_slugs: Should be of type str"
            )

        assert create_missing is not None and isinstance(create_missing, bool), (
            "create_missing shouldn't be null and should be of type bool"
        )

        if not self.corpus_types:
            self.list_corpus_types()

        for slug in type_slugs:
            # Do nothing if the type already exists
            if slug in self.corpus_types:
                continue

            # Do not create missing if not requested
            if not create_missing:
                raise MissingElementType(
                    f"Element type `{slug}` was not in the corpus."
                )

            # Create the type if non-existent
            self.create_element_type(slug=slug, name=slug)

    @unsupported_cache
    def create_sub_element(
        self,
        element: Element,
        type: str,
        name: str,
        polygon: list[list[int | float]] | None = None,
        confidence: float | None = None,
        image: str | None = None,
        slim_output: bool = True,
    ) -> str:
        """
        Create a child element on the given element through the API.

        :param Element element: The parent element.
        :param type: Slug of the element type for this child element.
        :param name: Name of the child element.
        :param polygon: Optional polygon of the child element.
        :param confidence: Optional confidence score, between 0.0 and 1.0.
        :param image: Optional image ID of the child element.
        :param slim_output: Whether to return the child ID or the full child.
        :returns: UUID of the created element.
        """
        assert element and isinstance(element, Element), (
            "element shouldn't be null and should be of type Element"
        )
        assert type and isinstance(type, str), (
            "type shouldn't be null and should be of type str"
        )
        assert name and isinstance(name, str), (
            "name shouldn't be null and should be of type str"
        )
        assert polygon is None or isinstance(polygon, list), (
            "polygon should be None or a list"
        )
        if polygon is not None:
            assert len(polygon) >= 3, "polygon should have at least three points"
            assert all(
                isinstance(point, list) and len(point) == 2 for point in polygon
            ), "polygon points should be lists of two items"
            assert all(
                isinstance(coord, int | float) for point in polygon for coord in point
            ), "polygon points should be lists of two numbers"
        assert confidence is None or (
            isinstance(confidence, float) and 0 <= confidence <= 1
        ), "confidence should be None or a float in [0..1] range"
        assert image is None or isinstance(image, str), "image should be None or string"
        if image is not None:
            # Make sure it's a valid UUID
            try:
                UUID(image)
            except ValueError as e:
                raise ValueError("image is not a valid uuid.") from e
        if polygon and image is None:
            assert element.zone, (
                "An image or a parent with an image is required to create an element with a polygon."
            )
        assert isinstance(slim_output, bool), "slim_output should be of type bool"

        if self.is_read_only:
            logger.warning("Cannot create element as this worker is in read-only mode")
            return

        sub_element = self.api_client.request(
            "CreateElement",
            body={
                "type": type,
                "name": name,
                "image": image,
                "corpus": element.corpus.id,
                "polygon": polygon,
                "parent": element.id,
                "worker_run_id": self.worker_run_id,
                "confidence": confidence,
            },
        )

        return sub_element["id"] if slim_output else sub_element

    @batch_publication
    def create_elements(
        self,
        parent: Element | CachedElement,
        elements: list[dict[str, str | list[list[int | float]] | float | None]],
        batch_size: int = DEFAULT_BATCH_SIZE,
    ) -> list[dict[str, str]]:
        """
        Create child elements on the given element in a single API request.

        :param parent: Parent element for all the new child elements. The parent must have an image and a polygon.
        :param elements: List of dicts, one per element. Each dict can have the following keys:

            name (str)
               Required. Name of the element.

            type (str)
               Required. Slug of the element type for this element.

            polygon (list(list(int or float)))
               Required. Polygon for this child element. Must have at least three points, with each point
               having two non-negative coordinates and being inside of the parent element's image.

            confidence (float or None)
                Optional confidence score, between 0.0 and 1.0.

        :param batch_size: The size of each batch, which will be used to split the publication to avoid API errors.

        :return: List of dicts, with each dict having a single key, ``id``, holding the UUID of each created element.
        """
        if isinstance(parent, Element):
            assert parent.get("zone"), (
                "create_elements cannot be used on parents without zones"
            )
        elif isinstance(parent, CachedElement):
            assert parent.image_id, (
                "create_elements cannot be used on parents without images"
            )
        else:
            raise TypeError(
                "Parent element should be an Element or CachedElement instance"
            )

        assert elements and isinstance(elements, list), (
            "elements shouldn't be null and should be of type list"
        )

        for index, element in enumerate(elements):
            assert isinstance(element, dict), (
                f"Element at index {index} in elements: Should be of type dict"
            )

            name = element.get("name")
            assert name and isinstance(name, str), (
                f"Element at index {index} in elements: name shouldn't be null and should be of type str"
            )

            type = element.get("type")
            assert type and isinstance(type, str), (
                f"Element at index {index} in elements: type shouldn't be null and should be of type str"
            )

            polygon = element.get("polygon")
            assert polygon and isinstance(polygon, list), (
                f"Element at index {index} in elements: polygon shouldn't be null and should be of type list"
            )
            assert len(polygon) >= 3, (
                f"Element at index {index} in elements: polygon should have at least three points"
            )
            assert all(
                isinstance(point, list) and len(point) == 2 for point in polygon
            ), (
                f"Element at index {index} in elements: polygon points should be lists of two items"
            )
            assert all(
                isinstance(coord, int | float) for point in polygon for coord in point
            ), (
                f"Element at index {index} in elements: polygon points should be lists of two numbers"
            )

            confidence = element.get("confidence")
            assert confidence is None or (
                isinstance(confidence, float) and 0 <= confidence <= 1
            ), (
                f"Element at index {index} in elements: confidence should be None or a float in [0..1] range"
            )

        if self.is_read_only:
            logger.warning("Cannot create elements as this worker is in read-only mode")
            return

        created_ids = [
            created_id
            for batch in make_batches(elements, "element", batch_size)
            for created_id in self.api_client.request(
                "CreateElements",
                id=parent.id,
                body={
                    "worker_run_id": self.worker_run_id,
                    "elements": batch,
                },
            )
        ]

        if self.use_cache:
            # Create the image as needed and handle both an Element and a CachedElement
            if isinstance(parent, CachedElement):
                image_id = parent.image_id
            else:
                image_id = parent.zone.image.id
                CachedImage.get_or_create(
                    id=parent.zone.image.id,
                    defaults={
                        "width": parent.zone.image.width,
                        "height": parent.zone.image.height,
                        "url": parent.zone.image.url,
                    },
                )

            # Store elements in local cache
            try:
                to_insert = [
                    {
                        "id": created_ids[idx]["id"],
                        "parent_id": parent.id,
                        "type": element["type"],
                        "image_id": image_id,
                        "polygon": element["polygon"],
                        "worker_run_id": self.worker_run_id,
                        "confidence": element.get("confidence"),
                    }
                    for idx, element in enumerate(elements)
                ]
                CachedElement.insert_many(to_insert).execute()
            except IntegrityError as e:
                logger.warning(f"Couldn't save created elements in local cache: {e}")

        return created_ids

    @unsupported_cache
    def create_element_parent(
        self,
        parent: Element,
        child: Element,
    ) -> dict[str, str]:
        """
        Link an element to a parent through the API.

        :param parent: Parent element.
        :param child: Child element.
        :returns: A dict from the ``CreateElementParent`` API endpoint.
        """
        assert parent and isinstance(parent, Element), (
            "parent shouldn't be null and should be of type Element"
        )
        assert child and isinstance(child, Element), (
            "child shouldn't be null and should be of type Element"
        )

        if self.is_read_only:
            logger.warning("Cannot link elements as this worker is in read-only mode")
            return

        return self.api_client.request(
            "CreateElementParent",
            parent=parent.id,
            child=child.id,
        )

    @unsupported_cache
    @batch_publication
    def create_element_children(
        self,
        parent: Element,
        children: list[Element],
        batch_size: int = DEFAULT_BATCH_SIZE,
    ) -> list[str]:
        """
        Link multiple elements to a single parent through the API.

        :param parent: Parent element.
        :param children: A list of child elements.
        :param batch_size: The size of each batch, which will be used to split the publication to avoid API errors.

        :returns: A list containing the string UUID of each child linked to the parent.
        """
        assert parent and isinstance(parent, Element), (
            "parent shouldn't be null and should be of type Element"
        )

        assert children and isinstance(children, list), (
            "children shouldn't be null and should be of type list"
        )

        for index, child in enumerate(children):
            assert isinstance(child, Element), (
                f"Child at index {index} in children: Should be of type Element"
            )

        if self.is_read_only:
            logger.warning("Cannot link elements as this worker is in read-only mode")
            return

        return [
            child_id
            for batch in make_batches(children, "child", batch_size)
            for child_id in self.api_client.request(
                "CreateElementChildren",
                id=parent.id,
                body={
                    "children": list(map(attrgetter("id"), batch)),
                },
            )["children"]
        ]

    def partial_update_element(
        self, element: Element | CachedElement, **kwargs
    ) -> dict:
        """
        Partially updates an element through the API.

        :param element: The element to update.
        :param **kwargs:

            * *type* (``str``): Optional slug type of the element.
            * *name* (``str``): Optional name of the element.
            * *polygon* (``list``): Optional polygon for this element
            * *confidence* (``float``): Optional confidence score of this element
            * *rotation_angle* (``int``): Optional rotation angle of this element
            * *mirrored* (``bool``): Optional mirror status of this element
            * *image* (``UUID``): Optional ID of the image of this element


        :returns: A dict from the ``PartialUpdateElement`` API endpoint.
        """
        assert element and isinstance(element, Element | CachedElement), (
            "element shouldn't be null and should be an Element or CachedElement"
        )

        if "type" in kwargs:
            assert isinstance(kwargs["type"], str), "type should be a str"

        if "name" in kwargs:
            assert isinstance(kwargs["name"], str), "name should be a str"

        if "polygon" in kwargs:
            polygon = kwargs["polygon"]
            assert isinstance(polygon, list), "polygon should be a list"
            assert len(polygon) >= 3, "polygon should have at least three points"
            assert all(
                isinstance(point, list) and len(point) == 2 for point in polygon
            ), "polygon points should be lists of two items"
            assert all(
                isinstance(coord, int | float) for point in polygon for coord in point
            ), "polygon points should be lists of two numbers"

        if "confidence" in kwargs:
            confidence = kwargs["confidence"]
            assert confidence is None or (
                isinstance(confidence, float) and 0 <= confidence <= 1
            ), "confidence should be None or a float in [0..1] range"

        if "rotation_angle" in kwargs:
            rotation_angle = kwargs["rotation_angle"]
            assert isinstance(rotation_angle, int) and rotation_angle >= 0, (
                "rotation_angle should be a positive integer"
            )

        if "mirrored" in kwargs:
            assert isinstance(kwargs["mirrored"], bool), "mirrored should be a boolean"

        if "image" in kwargs:
            image = kwargs["image"]
            assert isinstance(image, UUID), "image should be a UUID"
            # Cast to string
            kwargs["image"] = str(image)

        if self.is_read_only:
            logger.warning("Cannot update element as this worker is in read-only mode")
            return

        updated_element = self.api_client.request(
            "PartialUpdateElement",
            id=element.id,
            body=kwargs,
        )

        if self.use_cache:
            # Name is not present in CachedElement model
            kwargs.pop("name", None)

            # Stringify polygon if present
            if "polygon" in kwargs:
                kwargs["polygon"] = str(kwargs["polygon"])

            # Retrieve the right image
            if "image" in kwargs:
                kwargs["image"] = CachedImage.get_by_id(kwargs["image"])

            CachedElement.update(**kwargs).where(
                CachedElement.id == element.id
            ).execute()

        return updated_element

    def list_elements(
        self,
        folder: bool | None = None,
        name: str | None = None,
        top_level: bool | None = None,
        transcription_worker_version: str | bool | None = None,
        transcription_worker_run: str | bool | None = None,
        type: str | None = None,
        with_classes: bool | None = None,
        with_corpus: bool | None = None,
        with_metadata: bool | None = None,
        with_has_children: bool | None = None,
        with_zone: bool | None = None,
        worker_version: str | bool | None = None,
        worker_run: str | bool | None = None,
    ) -> Iterable[dict] | Iterable[CachedElement]:
        """
        List element in a corpus.

        Warns:
        ----
        The following parameters are **deprecated**:

        - `transcription_worker_version` in favor of `transcription_worker_run`
        - `worker_version` in favor of `worker_run`

        :param folder: Restrict to or exclude elements with folder types.
           This parameter is not supported when caching is enabled.
        :param name: Restrict to elements whose name contain a substring (case-insensitive).
           This parameter is not supported when caching is enabled.
        :param top_level: Restrict to or exclude folder elements without parent elements (top-level elements).
           This parameter is not supported when caching is enabled.
        :param transcription_worker_version: **Deprecated** Restrict to elements that have a transcription created by a worker version with this UUID. Set to False to look for elements that have a manual transcription.
           This parameter is not supported when caching is enabled.
        :param transcription_worker_run: Restrict to elements that have a transcription created by a worker run with this UUID. Set to False to look for elements that have a manual transcription.
           This parameter is not supported when caching is enabled.
        :param type: Restrict to elements with a specific type slug
           This parameter is not supported when caching is enabled.
        :param with_classes: Include each element's classifications in the response.
           This parameter is not supported when caching is enabled.
        :param with_corpus: Include each element's corpus in the response.
           This parameter is not supported when caching is enabled.
        :param with_has_children: Include the ``has_children`` attribute in the response,
           indicating if this element has child elements of its own.
           This parameter is not supported when caching is enabled.
        :param with_metadata: Include each element's metadata in the response.
           This parameter is not supported when caching is enabled.
        :param with_zone: Include the ``zone`` attribute in the response,
           holding the element's image and polygon.
           This parameter is not supported when caching is enabled.
        :param worker_version: **Deprecated** Restrict to elements created by a worker version with this UUID.
        :param worker_run: Restrict to elements created by a worker run with this UUID.
        :return: An iterable of dicts from the ``ListElementChildren`` API endpoint,
           or an iterable of [CachedElement][arkindex_worker.cache.CachedElement] when caching is enabled.
        """
        query_params = {}
        if folder is not None:
            assert isinstance(folder, bool), "folder should be of type bool"
            query_params["folder"] = folder
        if name:
            assert isinstance(name, str), "name should be of type str"
            query_params["name"] = name
        if top_level is not None:
            assert isinstance(top_level, bool), "top_level should be of type bool"
            query_params["top_level"] = top_level
        if transcription_worker_version is not None:
            warn(
                "`transcription_worker_version` usage is deprecated. Consider using `transcription_worker_run` instead.",
                DeprecationWarning,
                stacklevel=1,
            )
            assert isinstance(transcription_worker_version, str | bool), (
                "transcription_worker_version should be of type str or bool"
            )
            if isinstance(transcription_worker_version, bool):
                assert transcription_worker_version is False, (
                    "if of type bool, transcription_worker_version can only be set to False"
                )
            query_params["transcription_worker_version"] = transcription_worker_version
        if transcription_worker_run is not None:
            assert isinstance(transcription_worker_run, str | bool), (
                "transcription_worker_run should be of type str or bool"
            )
            if isinstance(transcription_worker_run, bool):
                assert transcription_worker_run is False, (
                    "if of type bool, transcription_worker_run can only be set to False"
                )
            query_params["transcription_worker_run"] = transcription_worker_run
        if type:
            assert isinstance(type, str), "type should be of type str"
            query_params["type"] = type
        if with_classes is not None:
            assert isinstance(with_classes, bool), "with_classes should be of type bool"
            query_params["with_classes"] = with_classes
        if with_corpus is not None:
            assert isinstance(with_corpus, bool), "with_corpus should be of type bool"
            query_params["with_corpus"] = with_corpus
        if with_has_children is not None:
            assert isinstance(with_has_children, bool), (
                "with_has_children should be of type bool"
            )
            query_params["with_has_children"] = with_has_children
        if with_metadata is not None:
            assert isinstance(with_metadata, bool), (
                "with_metadata should be of type bool"
            )
            query_params["with_metadata"] = with_metadata
        if with_zone is not None:
            assert isinstance(with_zone, bool), "with_zone should be of type bool"
            query_params["with_zone"] = with_zone
        if worker_version is not None:
            warn(
                "`worker_version` usage is deprecated. Consider using `worker_run` instead.",
                DeprecationWarning,
                stacklevel=1,
            )
            assert isinstance(worker_version, str | bool), (
                "worker_version should be of type str or bool"
            )
            if isinstance(worker_version, bool):
                assert worker_version is False, (
                    "if of type bool, worker_version can only be set to False"
                )
            query_params["worker_version"] = worker_version
        if worker_run is not None:
            assert isinstance(worker_run, str | bool), (
                "worker_run should be of type str or bool"
            )
            if isinstance(worker_run, bool):
                assert worker_run is False, (
                    "if of type bool, worker_run can only be set to False"
                )
            query_params["worker_run"] = worker_run

        if not self.use_cache:
            return self.api_client.paginate(
                "ListElements", corpus=self.corpus_id, **query_params
            )

        # Checking that we only received query_params handled by the cache
        assert set(query_params.keys()) <= {
            "type",
            "worker_version",
            "worker_run",
        }, (
            "When using the local cache, you can only filter by 'type' and/or 'worker_version' and/or 'worker_run'"
        )

        query = CachedElement.select()
        if type:
            query = query.where(CachedElement.type == type)
        if worker_version is not None:
            # If worker_version=False, filter by manual worker_version e.g. None
            worker_version_id = worker_version or None
            if worker_version_id:
                query = query.where(
                    CachedElement.worker_version_id == worker_version_id
                )
            else:
                query = query.where(CachedElement.worker_version_id.is_null())

        if worker_run is not None:
            # If worker_run=False, filter by manual worker_run e.g. None
            worker_run_id = worker_run or None
            if worker_run_id:
                query = query.where(CachedElement.worker_run_id == worker_run_id)
            else:
                query = query.where(CachedElement.worker_run_id.is_null())

        return query

    def list_element_children(
        self,
        element: Element | CachedElement,
        folder: bool | None = None,
        name: str | None = None,
        recursive: bool | None = None,
        transcription_worker_version: str | bool | None = None,
        transcription_worker_run: str | bool | None = None,
        type: str | None = None,
        with_classes: bool | None = None,
        with_corpus: bool | None = None,
        with_metadata: bool | None = None,
        with_has_children: bool | None = None,
        with_zone: bool | None = None,
        worker_version: str | bool | None = None,
        worker_run: str | bool | None = None,
    ) -> Iterable[dict] | Iterable[CachedElement]:
        """
        List children of an element.

        Warns:
        ----
        The following parameters are **deprecated**:

        - `transcription_worker_version` in favor of `transcription_worker_run`
        - `worker_version` in favor of `worker_run`

        :param element: Parent element to find children of.
        :param folder: Restrict to or exclude elements with folder types.
           This parameter is not supported when caching is enabled.
        :param name: Restrict to elements whose name contain a substring (case-insensitive).
           This parameter is not supported when caching is enabled.
        :param recursive: Look for elements recursively (grand-children, etc.)
           This parameter is not supported when caching is enabled.
        :param transcription_worker_version: **Deprecated** Restrict to elements that have a transcription created by a worker version with this UUID. Set to False to look for elements that have a manual transcription.
           This parameter is not supported when caching is enabled.
        :param transcription_worker_run: Restrict to elements that have a transcription created by a worker run with this UUID. Set to False to look for elements that have a manual transcription.
           This parameter is not supported when caching is enabled.
        :param type: Restrict to elements with a specific type slug
           This parameter is not supported when caching is enabled.
        :param with_classes: Include each element's classifications in the response.
           This parameter is not supported when caching is enabled.
        :param with_corpus: Include each element's corpus in the response.
           This parameter is not supported when caching is enabled.
        :param with_has_children: Include the ``has_children`` attribute in the response,
           indicating if this element has child elements of its own.
           This parameter is not supported when caching is enabled.
        :param with_metadata: Include each element's metadata in the response.
           This parameter is not supported when caching is enabled.
        :param with_zone: Include the ``zone`` attribute in the response,
           holding the element's image and polygon.
           This parameter is not supported when caching is enabled.
        :param worker_version: **Deprecated** Restrict to elements created by a worker version with this UUID.
        :param worker_run: Restrict to elements created by a worker run with this UUID.
        :return: An iterable of dicts from the ``ListElementChildren`` API endpoint,
           or an iterable of [CachedElement][arkindex_worker.cache.CachedElement] when caching is enabled.
        """
        assert element and isinstance(element, Element | CachedElement), (
            "element shouldn't be null and should be an Element or CachedElement"
        )
        query_params = {}
        if folder is not None:
            assert isinstance(folder, bool), "folder should be of type bool"
            query_params["folder"] = folder
        if name:
            assert isinstance(name, str), "name should be of type str"
            query_params["name"] = name
        if recursive is not None:
            assert isinstance(recursive, bool), "recursive should be of type bool"
            query_params["recursive"] = recursive
        if transcription_worker_version is not None:
            warn(
                "`transcription_worker_version` usage is deprecated. Consider using `transcription_worker_run` instead.",
                DeprecationWarning,
                stacklevel=1,
            )
            assert isinstance(transcription_worker_version, str | bool), (
                "transcription_worker_version should be of type str or bool"
            )
            if isinstance(transcription_worker_version, bool):
                assert transcription_worker_version is False, (
                    "if of type bool, transcription_worker_version can only be set to False"
                )
            query_params["transcription_worker_version"] = transcription_worker_version
        if transcription_worker_run is not None:
            assert isinstance(transcription_worker_run, str | bool), (
                "transcription_worker_run should be of type str or bool"
            )
            if isinstance(transcription_worker_run, bool):
                assert transcription_worker_run is False, (
                    "if of type bool, transcription_worker_run can only be set to False"
                )
            query_params["transcription_worker_run"] = transcription_worker_run
        if type:
            assert isinstance(type, str), "type should be of type str"
            query_params["type"] = type
        if with_classes is not None:
            assert isinstance(with_classes, bool), "with_classes should be of type bool"
            query_params["with_classes"] = with_classes
        if with_corpus is not None:
            assert isinstance(with_corpus, bool), "with_corpus should be of type bool"
            query_params["with_corpus"] = with_corpus
        if with_has_children is not None:
            assert isinstance(with_has_children, bool), (
                "with_has_children should be of type bool"
            )
            query_params["with_has_children"] = with_has_children
        if with_metadata is not None:
            assert isinstance(with_metadata, bool), (
                "with_metadata should be of type bool"
            )
            query_params["with_metadata"] = with_metadata
        if with_zone is not None:
            assert isinstance(with_zone, bool), "with_zone should be of type bool"
            query_params["with_zone"] = with_zone
        if worker_version is not None:
            warn(
                "`worker_version` usage is deprecated. Consider using `worker_run` instead.",
                DeprecationWarning,
                stacklevel=1,
            )
            assert isinstance(worker_version, str | bool), (
                "worker_version should be of type str or bool"
            )
            if isinstance(worker_version, bool):
                assert worker_version is False, (
                    "if of type bool, worker_version can only be set to False"
                )
            query_params["worker_version"] = worker_version
        if worker_run is not None:
            assert isinstance(worker_run, str | bool), (
                "worker_run should be of type str or bool"
            )
            if isinstance(worker_run, bool):
                assert worker_run is False, (
                    "if of type bool, worker_run can only be set to False"
                )
            query_params["worker_run"] = worker_run

        if not self.use_cache:
            return self.api_client.paginate(
                "ListElementChildren", id=element.id, **query_params
            )

        # Checking that we only received query_params handled by the cache
        assert set(query_params.keys()) <= {
            "type",
            "worker_version",
            "worker_run",
        }, (
            "When using the local cache, you can only filter by 'type' and/or 'worker_version' and/or 'worker_run'"
        )

        query = CachedElement.select().where(CachedElement.parent_id == element.id)
        if type:
            query = query.where(CachedElement.type == type)
        if worker_version is not None:
            # If worker_version=False, filter by manual worker_version e.g. None
            worker_version_id = worker_version or None
            if worker_version_id:
                query = query.where(
                    CachedElement.worker_version_id == worker_version_id
                )
            else:
                query = query.where(CachedElement.worker_version_id.is_null())

        if worker_run is not None:
            # If worker_run=False, filter by manual worker_run e.g. None
            worker_run_id = worker_run or None
            if worker_run_id:
                query = query.where(CachedElement.worker_run_id == worker_run_id)
            else:
                query = query.where(CachedElement.worker_run_id.is_null())

        return query

    def list_element_parents(
        self,
        element: Element | CachedElement,
        folder: bool | None = None,
        name: str | None = None,
        recursive: bool | None = None,
        transcription_worker_version: str | bool | None = None,
        transcription_worker_run: str | bool | None = None,
        type: str | None = None,
        with_classes: bool | None = None,
        with_corpus: bool | None = None,
        with_metadata: bool | None = None,
        with_has_children: bool | None = None,
        with_zone: bool | None = None,
        worker_version: str | bool | None = None,
        worker_run: str | bool | None = None,
    ) -> Iterable[dict] | Iterable[CachedElement]:
        """
        List parents of an element.

        Warns:
        ----
        The following parameters are **deprecated**:

        - `transcription_worker_version` in favor of `transcription_worker_run`
        - `worker_version` in favor of `worker_run`

        :param element: Child element to find parents of.
        :param folder: Restrict to or exclude elements with folder types.
           This parameter is not supported when caching is enabled.
        :param name: Restrict to elements whose name contain a substring (case-insensitive).
           This parameter is not supported when caching is enabled.
        :param recursive: Look for elements recursively (grand-children, etc.)
           This parameter is not supported when caching is enabled.
        :param transcription_worker_version: **Deprecated** Restrict to elements that have a transcription created by a worker version with this UUID.
           This parameter is not supported when caching is enabled.
        :param transcription_worker_run: Restrict to elements that have a transcription created by a worker run with this UUID.
           This parameter is not supported when caching is enabled.
        :param type: Restrict to elements with a specific type slug
           This parameter is not supported when caching is enabled.
        :param with_classes: Include each element's classifications in the response.
           This parameter is not supported when caching is enabled.
        :param with_corpus: Include each element's corpus in the response.
           This parameter is not supported when caching is enabled.
        :param with_has_children: Include the ``has_children`` attribute in the response,
           indicating if this element has child elements of its own.
           This parameter is not supported when caching is enabled.
        :param with_metadata: Include each element's metadata in the response.
           This parameter is not supported when caching is enabled.
        :param with_zone: Include the ``zone`` attribute in the response,
           holding the element's image and polygon.
           This parameter is not supported when caching is enabled.
        :param worker_version: **Deprecated** Restrict to elements created by a worker version with this UUID.
        :param worker_run: Restrict to elements created by a worker run with this UUID.
        :return: An iterable of dicts from the ``ListElementParents`` API endpoint,
           or an iterable of [CachedElement][arkindex_worker.cache.CachedElement] when caching is enabled.
        """
        assert element and isinstance(element, Element | CachedElement), (
            "element shouldn't be null and should be an Element or CachedElement"
        )
        query_params = {}
        if folder is not None:
            assert isinstance(folder, bool), "folder should be of type bool"
            query_params["folder"] = folder
        if name:
            assert isinstance(name, str), "name should be of type str"
            query_params["name"] = name
        if recursive is not None:
            assert isinstance(recursive, bool), "recursive should be of type bool"
            query_params["recursive"] = recursive
        if transcription_worker_version is not None:
            warn(
                "`transcription_worker_version` usage is deprecated. Consider using `transcription_worker_run` instead.",
                DeprecationWarning,
                stacklevel=1,
            )
            assert isinstance(transcription_worker_version, str | bool), (
                "transcription_worker_version should be of type str or bool"
            )
            if isinstance(transcription_worker_version, bool):
                assert transcription_worker_version is False, (
                    "if of type bool, transcription_worker_version can only be set to False"
                )
            query_params["transcription_worker_version"] = transcription_worker_version
        if transcription_worker_run is not None:
            assert isinstance(transcription_worker_run, str | bool), (
                "transcription_worker_run should be of type str or bool"
            )
            if isinstance(transcription_worker_run, bool):
                assert transcription_worker_run is False, (
                    "if of type bool, transcription_worker_run can only be set to False"
                )
            query_params["transcription_worker_run"] = transcription_worker_run
        if type:
            assert isinstance(type, str), "type should be of type str"
            query_params["type"] = type
        if with_classes is not None:
            assert isinstance(with_classes, bool), "with_classes should be of type bool"
            query_params["with_classes"] = with_classes
        if with_corpus is not None:
            assert isinstance(with_corpus, bool), "with_corpus should be of type bool"
            query_params["with_corpus"] = with_corpus
        if with_has_children is not None:
            assert isinstance(with_has_children, bool), (
                "with_has_children should be of type bool"
            )
            query_params["with_has_children"] = with_has_children
        if with_metadata is not None:
            assert isinstance(with_metadata, bool), (
                "with_metadata should be of type bool"
            )
            query_params["with_metadata"] = with_metadata
        if with_zone is not None:
            assert isinstance(with_zone, bool), "with_zone should be of type bool"
            query_params["with_zone"] = with_zone
        if worker_version is not None:
            warn(
                "`worker_version` usage is deprecated. Consider using `worker_run` instead.",
                DeprecationWarning,
                stacklevel=1,
            )
            assert isinstance(worker_version, str | bool), (
                "worker_version should be of type str or bool"
            )
            if isinstance(worker_version, bool):
                assert worker_version is False, (
                    "if of type bool, worker_version can only be set to False"
                )
            query_params["worker_version"] = worker_version
        if worker_run is not None:
            assert isinstance(worker_run, str | bool), (
                "worker_run should be of type str or bool"
            )
            if isinstance(worker_run, bool):
                assert worker_run is False, (
                    "if of type bool, worker_run can only be set to False"
                )
            query_params["worker_run"] = worker_run

        if not self.use_cache:
            return self.api_client.paginate(
                "ListElementParents", id=element.id, **query_params
            )

        # Checking that we only received query_params handled by the cache
        assert set(query_params.keys()) <= {
            "type",
            "worker_version",
            "worker_run",
        }, (
            "When using the local cache, you can only filter by 'type' and/or 'worker_version' and/or 'worker_run'"
        )

        parent_ids = CachedElement.select(CachedElement.parent_id).where(
            CachedElement.id == element.id
        )
        query = CachedElement.select().where(CachedElement.id.in_(parent_ids))
        if type:
            query = query.where(CachedElement.type == type)
        if worker_version is not None:
            # If worker_version=False, filter by manual worker_version e.g. None
            worker_version_id = worker_version or None
            if worker_version_id:
                query = query.where(
                    CachedElement.worker_version_id == worker_version_id
                )
            else:
                query = query.where(CachedElement.worker_version_id.is_null())

        if worker_run is not None:
            # If worker_run=False, filter by manual worker_run e.g. None
            worker_run_id = worker_run or None
            if worker_run_id:
                query = query.where(CachedElement.worker_run_id == worker_run_id)
            else:
                query = query.where(CachedElement.worker_run_id.is_null())

        return query



"""
The base class for all Arkindex workers.
"""

import argparse
import json
import logging
import os
import shutil
from pathlib import Path
from tempfile import mkdtemp
from typing import Any

import gnupg
import yaml

from arkindex import options_from_env
from arkindex.exceptions import ErrorResponse
from arkindex_worker import logger
from arkindex_worker.cache import (
    check_version,
    create_tables,
    create_version_table,
    init_cache_db,
    merge_parents_cache,
)
from arkindex_worker.utils import close_delete_file, extract_tar_zst_archive
from arkindex_worker.worker.process import ProcessMode
from teklia_toolbox.requests import get_arkindex_client


class ExtrasDirNotFoundError(Exception):
    """
    Exception raised when the path towards the extras directory is invalid
    """


class BaseWorker:
    """
    Base class for Arkindex workers.
    """

    def __init__(
        self,
        description: str | None = "Arkindex Base Worker",
        support_cache: bool | None = False,
    ):
        """
        Initialize the worker.

        :param description: Description shown in the ``worker-...`` command line tool.
        :param support_cache: Whether or not this worker supports the cache database.
           Override the constructor and set this parameter to start using the cache database.
        """

        self.parser = argparse.ArgumentParser(description=description)
        self.parser.add_argument(
            "-c",
            "--config",
            help="Alternative configuration file when running without a Worker Run ID",
            type=open,
        )
        self.parser.add_argument(
            "-d",
            "--database",
            help="Alternative SQLite database to use for worker caching",
            type=Path,
            default=None,
        )
        self.parser.add_argument(
            "-v",
            "--verbose",
            "--debug",
            help="Display more information on events and errors",
            action="store_true",
            default=False,
        )
        self.parser.add_argument(
            "--dev",
            help=(
                "Run worker in developer mode. "
                "Worker will be in read-only state even if a worker run is supplied. "
            ),
            action="store_true",
            default=False,
        )
        # To load models, datasets, etc, locally
        self.parser.add_argument(
            "--extras-dir",
            help=(
                "The path to a local directory to store extra files like models, datasets, etc (development only)."
            ),
            type=Path,
        )

        # Call potential extra arguments
        self.add_arguments()

        # Setup workdir either in Ponos environment or on host's home
        if os.environ.get("PONOS_DATA"):
            self.work_dir = Path(os.environ["PONOS_DATA"], "current")
        else:
            # We use the official XDG convention to store file for developers
            # https://specifications.freedesktop.org/basedir-spec/basedir-spec-latest.html
            xdg_data_home = os.environ.get("XDG_DATA_HOME", "~/.local/share")
            self.work_dir = Path(xdg_data_home, "arkindex").expanduser()
            self.work_dir.mkdir(parents=True, exist_ok=True)

        # Store task ID and chunk index. This is only available when running in production
        # through a ponos agent
        self.task_id = os.environ.get("PONOS_TASK")
        self.task_chunk = os.environ.get("ARKINDEX_TASK_CHUNK")

        # Store task data directory.
        self.task_data_dir = Path(os.environ.get("PONOS_DATA", "/data"))

        self.worker_run_id = os.environ.get("ARKINDEX_WORKER_RUN_ID")
        if not self.worker_run_id:
            logger.warning(
                "Missing ARKINDEX_WORKER_RUN_ID environment variable, worker is in read-only mode"
            )

        logger.info(f"Worker will use {self.work_dir} as working directory")

        self.process_information = None
        # corpus_id will be updated in configure() using the worker_run's corpus
        # or in configure_for_developers() from the environment
        self._corpus_id = None
        self.user_configuration = {}
        self.model_configuration = {}
        self.support_cache = support_cache
        # use_cache will be updated in configure_cache() if the cache is supported and if
        # there is at least one available sqlite database either given or in the parent tasks
        self.use_cache = False

        # model_version_id will be updated in configure() using the worker_run's model version
        # or in configure_for_developers() from the environment
        self.model_version_id = None
        # model_details will be updated in configure() using the worker_run's model version
        # or in configure_for_developers() from the environment
        self.model_details = {}

        # task_parents will be updated in configure_cache() if the cache is supported,
        # if the task ID is set and if no database is passed as argument
        self.task_parents = []

        # Define API Client
        self.setup_api_client()

        # Known and available classes in processed corpus
        self.classes = {}
        # Known and available entity types in processed corpus
        self.entity_types = {}
        # Known and available element types in processed corpus
        self.corpus_types = {}

    @property
    def corpus_id(self) -> str:
        """
        ID of the corpus on which the worker is executed.
        Has to be set through the `ARKINDEX_CORPUS_ID` variable in **read-only** mode.
        Raises an Exception when trying to access when unset.
        """
        if not self._corpus_id:
            raise Exception("Missing ARKINDEX_CORPUS_ID environment variable")
        return self._corpus_id

    @property
    def process_mode(self) -> ProcessMode | None:
        """Mode of the process being run. Returns None when read-only."""
        if self.is_read_only:
            return
        return ProcessMode(self.process_information["mode"])

    @property
    def is_read_only(self) -> bool:
        """
        Whether or not the worker can publish data.

        False when dev mode is enabled with the ``--dev`` CLI argument,
            when no worker run ID is provided
        """
        return self.args.dev or self.worker_run_id is None

    @property
    def worker_version_id(self):
        """Deprecated property previously used to retrieve the current WorkerVersion ID.

        :raises DeprecationWarning: Whenever `worker_version_id` is used.
        """
        raise DeprecationWarning("`worker_version_id` usage is deprecated")

    def setup_api_client(self):
        """
        Create an ArkindexClient to make API requests towards Arkindex instances.
        """
        # Build Arkindex API client from environment variables
        self.api_client = get_arkindex_client(**options_from_env())
        logger.debug(f"Setup Arkindex API client on {self.api_client.document.url}")

    def configure_for_developers(self):
        """
        Setup the necessary configuration needed when working in `read_only` mode.
        This is the method called when running a worker locally.
        """
        assert self.is_read_only
        # Setup logging level if verbose or if ARKINDEX_DEBUG is set to true
        if self.args.verbose or os.environ.get("ARKINDEX_DEBUG"):
            logger.setLevel(logging.DEBUG)
            logger.debug("Debug output enabled")

        if self.args.config:
            # Load config from YAML file
            self.config = yaml.safe_load(self.args.config)
            self.worker_details = {"name": "Local worker"}
            required_secrets = self.config.get("secrets", [])
            logger.info(
                f"Running with local configuration from {self.args.config.name}"
            )
        else:
            self.config = {}
            self.worker_details = {}
            required_secrets = []
            logger.warning("Running without any extra configuration")

        # Define corpus_id from environment
        self._corpus_id = os.environ.get("ARKINDEX_CORPUS_ID")

        # Define model_version_id from environment
        self.model_version_id = os.environ.get("ARKINDEX_MODEL_VERSION_ID")

        # Define model_details from environment
        self.model_details = {"id": os.environ.get("ARKINDEX_MODEL_ID")}

        # Load all required secrets
        self.secrets = {name: self.load_secret(Path(name)) for name in required_secrets}

    def configure_worker_run(self):
        """
        Setup the necessary configuration needed using CLI args and environment variables.
        This is the method called when running a worker on Arkindex.
        """
        assert not self.is_read_only
        # Setup logging level if verbose or if ARKINDEX_DEBUG is set to true
        if self.args.verbose or os.environ.get("ARKINDEX_DEBUG"):
            logger.setLevel(logging.DEBUG)
            logger.debug("Debug output enabled")

        # Load worker run information
        worker_run = self.api_client.request("RetrieveWorkerRun", id=self.worker_run_id)

        # Load process information
        self.process_information = worker_run["process"]

        # Load corpus id
        self._corpus_id = worker_run["process"]["corpus"]

        # Load worker version information
        worker_version = worker_run["worker_version"]
        self.worker_details = worker_version["worker"]

        logger.info(f"Loaded {worker_run['summary']} from API")

        # The `RetrieveSecret` endpoint is only available in Arkindex EE.
        # In CE, the values of `secret` fields should be used directly without calling `RetrieveSecret`.
        can_retrieve_secret = "RetrieveSecret" in self.api_client.document.links

        def _process_config_item(item: dict) -> tuple[str, Any]:
            if not item["secret"]:
                return (item["key"], item["value"])

            # The secret may not be picked by the user
            if item["value"] is None:
                logger.info(f"Optional secret `{item['key']}` is not set")
                return (item["key"], None)

            value = item["value"]
            # Load secret when `RetrieveSecret` is available
            if can_retrieve_secret:
                value = self.load_secret(Path(item["value"]))

            return (item["key"], value)

        # Load model version configuration when available
        # Workers will use model version ID and details to download the model
        model_version = worker_run.get("model_version")
        if model_version:
            logger.info("Loaded model version configuration from WorkerRun")
            self.model_configuration.update(model_version["configuration"])

            # Set model_version ID as worker attribute
            self.model_version_id = model_version["id"]

            # Set model details as worker attribute
            self.model_details = model_version["model"]

        # Load worker run information
        try:
            config = self.api_client.request(
                "RetrieveWorkerRunConfiguration", id=self.worker_run_id
            )

            # Provide the same configuration through all previous attributes
            self.config = self.user_configuration = dict(
                map(_process_config_item, config["configuration"])
            )

            # Provide secret values through the previous attribute
            self.secrets = {
                item["key"]: self.config[item["key"]]
                for item in config["configuration"]
                if item["secret"]
            }
            logger.info("Using modern configuration")

            # Reset the model configuration to make sure workers rely on the single new source
            self.model_configuration = {}

            return  # Stop here once we have modern configuration

        except ErrorResponse as e:
            if e.status_code != 400:
                raise
            logger.info("Modern configuration is not available")

        # Use old-style configuration with local merge
        # Retrieve initial configuration from API
        self.config = worker_version["configuration"].get("configuration", {})
        if "user_configuration" in worker_version["configuration"]:
            # Add missing values (using the provided default if set) to user_configuration
            for key, value in worker_version["configuration"][
                "user_configuration"
            ].items():
                if key not in self.model_configuration:
                    self.user_configuration[key] = value.get("default")

        # Load all required secrets
        required_secrets = worker_version["configuration"].get("secrets", [])
        self.secrets = {name: self.load_secret(Path(name)) for name in required_secrets}

        # Load worker run configuration when available
        worker_configuration = worker_run.get("configuration")
        if worker_configuration and worker_configuration.get("configuration"):
            logger.info("Loaded user configuration from WorkerRun")
            self.user_configuration.update(worker_configuration.get("configuration"))

        # if debug mode is set to true activate debug mode in logger
        if self.user_configuration.get("debug"):
            logger.setLevel(logging.DEBUG)
            logger.debug("Debug output enabled")

    def configure_cache(self):
        """
        Setup the necessary attribute when using the cache system of `Base-Worker`.
        """
        paths = None
        if self.support_cache and self.args.database is not None:
            self.use_cache = True
        elif self.support_cache and self.task_id:
            task = self.api_client.request("RetrieveTask", id=self.task_id)
            self.task_parents = task["parents"]
            paths = self.find_parents_file_paths(Path("db.sqlite"))
            self.use_cache = len(paths) > 0

        if self.use_cache:
            if self.args.database is not None:
                assert self.args.database.is_file(), (
                    f"Database in {self.args.database} does not exist"
                )
                self.cache_path = self.args.database
            else:
                cache_dir = self.task_data_dir / self.task_id
                assert cache_dir.is_dir(), f"Missing task cache in {cache_dir}"
                self.cache_path = cache_dir / "db.sqlite"
            init_cache_db(self.cache_path)

            if self.args.database is not None:
                check_version(self.cache_path)
            else:
                create_version_table()

            create_tables()

            # Merging parents caches (if there are any) in the current task local cache, unless the database got overridden
            if self.args.database is None and paths is not None:
                merge_parents_cache(paths, self.cache_path)
        else:
            logger.debug("Cache is disabled")

    def configure(self):
        """
        Setup the worker using CLI arguments and environment variables.
        """
        # CLI args are stored on the instance so that implementations can access them
        self.args = self.parser.parse_args()

        if self.is_read_only:
            self.configure_for_developers()
        else:
            self.configure_worker_run()
            self.configure_cache()

        # Retrieve the model configuration
        if self.model_configuration:
            self.config.update(self.model_configuration)
            logger.info("Model version configuration retrieved")

        # Retrieve the user configuration
        if self.user_configuration:
            self.config.update(self.user_configuration)
            logger.info("User configuration retrieved")

    def load_secret(self, name: Path):
        """
        Load a Ponos secret by name.

        :param name: Name of the Ponos secret.
        :raises Exception: When the secret cannot be loaded from the API nor the local secrets directory.
        """
        secret = None

        # Load from the backend
        try:
            resp = self.api_client.request("RetrieveSecret", name=str(name))
            secret = resp["content"]
            logging.info(f"Loaded API secret {name}")
        except ErrorResponse as e:
            logger.warning(f"Secret {name} not available: {e.content}")

        # Load from local developer storage
        base_dir = Path(os.environ.get("XDG_CONFIG_HOME") or "~/.config").expanduser()
        path = base_dir / "arkindex" / "secrets" / name
        if path.exists():
            logging.debug(f"Loading local secret from {path}")

            try:
                gpg = gnupg.GPG()
                with path.open("rb") as gpg_file:
                    decrypted = gpg.decrypt_file(gpg_file)
                assert decrypted.ok, (
                    f"GPG error: {decrypted.status} - {decrypted.stderr}"
                )
                secret = decrypted.data.decode("utf-8")
                logging.info(f"Loaded local secret {name}")
            except Exception as e:
                logger.error(f"Local secret {name} is not available as {path}: {e}")

        if secret is None:
            raise Exception(f"Secret {name} is not available on the API nor locally")

        # Parse secret payload, according to its extension
        try:
            ext = name.suffix.lower()
            if ext == ".json":
                return json.loads(secret)
            elif ext in (".yaml", ".yml"):
                return yaml.safe_load(secret)
        except Exception as e:
            logger.error(f"Failed to parse secret {name}: {e}")

        # By default give raw secret payload
        return secret

    def find_extras_directory(self) -> Path:
        """
        Find the local path to the directory to store extra files. This supports two modes:
        - the worker runs in ponos, the directory is available at `/data/extra_files` (first try) or `/data/current`.
        - the worker runs locally, the developer may specify it using either
           - the `extras_dir` configuration parameter
           - the `--extras-dir` CLI parameter

        :return: Path to the directory for extra files on disk
        """
        if self.task_id:
            # When running in production with ponos, the agent
            # downloads the model and set it either in
            # - `/data/extra_files`
            # - the current task work dir
            extras_dir = self.task_data_dir / "extra_files"
            if extras_dir.exists():
                return extras_dir
            return self.work_dir
        else:
            extras_dir = self.config.get("extras_dir", self.args.extras_dir)
            if extras_dir is None:
                raise ExtrasDirNotFoundError(
                    "No path to the directory for extra files was provided. "
                    "Please provide extras_dir either through configuration "
                    "or as CLI argument."
                )
            extras_dir = Path(extras_dir)
            if not extras_dir.exists():
                raise ExtrasDirNotFoundError(
                    f"The path {extras_dir} does not link to any directory"
                )
            return extras_dir

    def find_parents_file_paths(self, filename: Path) -> list[Path]:
        """
        Find the paths of a specific file from the parent tasks.
        Only works if the task_parents attributes is updated, so if the cache is supported,
        if the task ID is set and if no database is passed as argument

        :param filename: Name of the file to find
        :return: Paths to the parent files
        """
        # Handle possible chunk in parent task name
        # This is needed to support the init_elements databases
        filenames = [
            filename,
        ]
        if self.task_chunk is not None:
            filenames.append(
                f"{filename.name.replace(filename.suffix, '')}_{self.task_chunk}{filename.suffix}"
            )

        # Find all the paths for these files
        return list(
            filter(
                lambda p: p.is_file(),
                [
                    self.task_data_dir / parent_id / name
                    for parent_id in self.task_parents
                    for name in filenames
                ],
            )
        )

    def extract_parent_archives(self, archive_name: Path, dest: Path) -> None:
        """
        Find and extract the paths from a specific file from the parent tasks.
        Only works if the task_parents attributes is updated, so if the cache is supported,
        if the task ID is set and if no database is passed as argument

        :param archive_name: Name of the file to find
        :param dest: Folder to store the extracted files
        """
        base_extracted_path = Path(mkdtemp(suffix="-extracted-data"))
        file_paths = self.find_parents_file_paths(archive_name)

        # Uncompress and extract the archive
        for file_path in file_paths:
            archive_fd, archive_path = extract_tar_zst_archive(
                file_path, base_extracted_path
            )
            # Remove the tar archive
            close_delete_file(archive_fd, archive_path)

            # Move all files in the dest folder
            for tmp_extracted_path in base_extracted_path.rglob("*"):
                if not tmp_extracted_path.is_file():
                    continue

                extracted_file = Path(
                    str(tmp_extracted_path).replace(str(base_extracted_path), str(dest))
                )
                extracted_file.parent.mkdir(parents=True, exist_ok=True)
                # Use shutil to avoid errors when the files are not on the same filesystem
                shutil.move(tmp_extracted_path, extracted_file)

            # Clean up
            shutil.rmtree(base_extracted_path)

    def add_arguments(self):
        """Override this method to add ``argparse`` arguments to this worker"""

    def run(self):
        """Override this method to implement your own process"""



"""
BaseWorker methods for tasks.
"""

import uuid
from collections.abc import Iterator
from http.client import REQUEST_TIMEOUT
from pathlib import Path

import magic
import requests

from arkindex.compat import DownloadedFile
from arkindex_worker import logger
from arkindex_worker.models import Artifact
from teklia_toolbox.requests import should_verify_cert


class TaskMixin:
    def list_artifacts(self, task_id: uuid.UUID) -> Iterator[Artifact]:
        """
        List artifacts associated to a task.

        :param task_id: Task ID to find artifacts from.
        :returns: An iterator of ``Artifact`` objects built from the ``ListArtifacts`` API endpoint.
        """
        assert task_id and isinstance(task_id, uuid.UUID), (
            "task_id shouldn't be null and should be an UUID"
        )

        results = self.api_client.request("ListArtifacts", id=task_id)

        return map(Artifact, results)

    def download_artifact(
        self, task_id: uuid.UUID, artifact: Artifact
    ) -> DownloadedFile:
        """
        Download an artifact content.

        :param task_id: Task ID the Artifact is from.
        :param artifact: Artifact to download content from.
        :returns: A temporary file containing the ``Artifact`` downloaded from the ``DownloadArtifact`` API endpoint.
        """
        assert task_id and isinstance(task_id, uuid.UUID), (
            "task_id shouldn't be null and should be an UUID"
        )
        assert artifact and isinstance(artifact, Artifact), (
            "artifact shouldn't be null and should be an Artifact"
        )

        return self.api_client.request(
            "DownloadArtifact", id=task_id, path=artifact.path
        )

    def upload_artifact(self, path: Path) -> None:
        """
        Upload a single file as an Artifact of the current task.

        :param path: Path of the single file to upload as an Artifact.
        """
        assert path and isinstance(path, Path) and path.exists(), (
            "path shouldn't be null, should be a Path and should exist"
        )

        if self.is_read_only:
            logger.warning("Cannot upload artifact as this worker is in read-only mode")
            return

        # Get path relative to task's data directory
        relpath = str(path.relative_to(self.work_dir))

        # Get file size
        size = path.stat().st_size

        # Detect content type
        try:
            content_type = magic.from_file(path, mime=True)
        except Exception as e:
            logger.warning(f"Failed to get a mime type for {path}: {e}")
            content_type = "application/octet-stream"

        # Create artifact on API to get an S3 url
        artifact = self.api_client.request(
            "CreateArtifact",
            id=self.task_id,
            body={"path": relpath, "content_type": content_type, "size": size},
        )

        # Upload the file content to S3
        s3_put_url = artifact["s3_put_url"]
        with path.open("rb") as content:
            resp = requests.put(
                s3_put_url,
                data=content,
                headers={"Content-Type": content_type},
                timeout=REQUEST_TIMEOUT,
                verify=should_verify_cert(s3_put_url),
            )
            resp.raise_for_status()



from collections.abc import Iterator
from enum import Enum

from arkindex_worker.cache import unsupported_cache

# Increases the number of elements returned per page by the API
PROCESS_ELEMENTS_PAGE_SIZE = 500


class ActivityState(Enum):
    """
    Processing state of an element.
    """

    Queued = "queued"
    """
    The element has not yet been processed by a worker.
    """

    Started = "started"
    """
    The element is being processed by a worker.
    """

    Processed = "processed"
    """
    The element has been successfully processed by a worker.
    """

    Error = "error"
    """
    An error occurred while processing this element.
    """


class ProcessMode(Enum):
    """
    Mode of the process of the worker.
    """

    Files = "files"
    """
    Processes of files (images, PDFs, IIIF, ...) imports.
    """

    Workers = "workers"
    """
    Processes of worker executions.
    """

    Template = "template"
    """
    Process templates.
    """

    S3 = "s3"
    """
    Processes of imports from an S3-compatible storage.
    """

    Local = "local"
    """
    Local processes.
    """

    Dataset = "dataset"
    """
    Dataset processes.
    """

    Export = "export"
    """
    Export processes.
    """


class ProcessMixin:
    @unsupported_cache
    def list_process_elements(self, with_image: bool = False) -> Iterator[dict]:
        """
        List the elements of a process.

        :param with_image: whether or not to include zone and image information in the elements response.
        :returns: the process' elements.
        """
        return self.api_client.paginate(
            "ListProcessElements",
            id=self.process_information["id"],
            with_image=with_image,
            allow_missing_data=True,
            page_size=PROCESS_ELEMENTS_PAGE_SIZE,
        )



import json
import re
from uuid import UUID

import pytest
from playhouse.shortcuts import model_to_dict

from arkindex.exceptions import ErrorResponse
from arkindex_worker.cache import CachedElement, CachedTranscription
from arkindex_worker.models import Element
from arkindex_worker.utils import DEFAULT_BATCH_SIZE
from arkindex_worker.worker.transcription import TextOrientation

from . import BASE_API_CALLS


def test_create_transcription_wrong_element(mock_elements_worker):
    with pytest.raises(
        AssertionError,
        match="element shouldn't be null and should be an Element or CachedElement",
    ):
        mock_elements_worker.create_transcription(
            element=None,
            text="i am a line",
            confidence=0.42,
        )

    with pytest.raises(
        AssertionError,
        match="element shouldn't be null and should be an Element or CachedElement",
    ):
        mock_elements_worker.create_transcription(
            element="not element type",
            text="i am a line",
            confidence=0.42,
        )


def test_create_transcription_wrong_text(mock_elements_worker):
    elt = Element({"id": "12341234-1234-1234-1234-123412341234"})

    with pytest.raises(
        AssertionError, match="text shouldn't be null and should be of type str"
    ):
        mock_elements_worker.create_transcription(
            element=elt,
            text=None,
            confidence=0.42,
        )

    with pytest.raises(
        AssertionError, match="text shouldn't be null and should be of type str"
    ):
        mock_elements_worker.create_transcription(
            element=elt,
            text=1234,
            confidence=0.42,
        )


def test_create_transcription_wrong_confidence(mock_elements_worker):
    elt = Element({"id": "12341234-1234-1234-1234-123412341234"})

    with pytest.raises(
        AssertionError,
        match=re.escape(
            "confidence shouldn't be null and should be a float in [0..1] range"
        ),
    ):
        mock_elements_worker.create_transcription(
            element=elt,
            text="i am a line",
            confidence=None,
        )

    with pytest.raises(
        AssertionError,
        match=re.escape(
            "confidence shouldn't be null and should be a float in [0..1] range"
        ),
    ):
        mock_elements_worker.create_transcription(
            element=elt,
            text="i am a line",
            confidence="wrong confidence",
        )

    with pytest.raises(
        AssertionError,
        match=re.escape(
            "confidence shouldn't be null and should be a float in [0..1] range"
        ),
    ):
        mock_elements_worker.create_transcription(
            element=elt,
            text="i am a line",
            confidence=0,
        )

    with pytest.raises(
        AssertionError,
        match=re.escape(
            "confidence shouldn't be null and should be a float in [0..1] range"
        ),
    ):
        mock_elements_worker.create_transcription(
            element=elt,
            text="i am a line",
            confidence=2.00,
        )


def test_create_transcription_default_orientation(responses, mock_elements_worker):
    elt = Element({"id": "12341234-1234-1234-1234-123412341234"})
    responses.add(
        responses.POST,
        f"http://testserver/api/v1/element/{elt.id}/transcription/",
        status=200,
        json={
            "id": "56785678-5678-5678-5678-567856785678",
            "text": "Animula vagula blandula",
            "confidence": 0.42,
            "worker_run_id": "56785678-5678-5678-5678-567856785678",
        },
    )
    mock_elements_worker.create_transcription(
        element=elt,
        text="Animula vagula blandula",
        confidence=0.42,
    )
    assert json.loads(responses.calls[-1].request.body) == {
        "text": "Animula vagula blandula",
        "worker_run_id": "56785678-5678-5678-5678-567856785678",
        "confidence": 0.42,
        "orientation": "horizontal-lr",
    }


def test_create_transcription_orientation(responses, mock_elements_worker):
    elt = Element({"id": "12341234-1234-1234-1234-123412341234"})
    responses.add(
        responses.POST,
        f"http://testserver/api/v1/element/{elt.id}/transcription/",
        status=200,
        json={
            "id": "56785678-5678-5678-5678-567856785678",
            "text": "Animula vagula blandula",
            "confidence": 0.42,
            "worker_run_id": "56785678-5678-5678-5678-567856785678",
        },
    )
    mock_elements_worker.create_transcription(
        element=elt,
        text="Animula vagula blandula",
        orientation=TextOrientation.VerticalLeftToRight,
        confidence=0.42,
    )
    assert json.loads(responses.calls[-1].request.body) == {
        "text": "Animula vagula blandula",
        "worker_run_id": "56785678-5678-5678-5678-567856785678",
        "confidence": 0.42,
        "orientation": "vertical-lr",
    }


def test_create_transcription_wrong_orientation(mock_elements_worker):
    elt = Element({"id": "12341234-1234-1234-1234-123412341234"})
    with pytest.raises(
        AssertionError,
        match="orientation shouldn't be null and should be of type TextOrientation",
    ):
        mock_elements_worker.create_transcription(
            element=elt,
            text="Animula vagula blandula",
            confidence=0.26,
            orientation="elliptical",
        )


def test_create_transcription_api_error(responses, mock_elements_worker):
    elt = Element({"id": "12341234-1234-1234-1234-123412341234"})
    responses.add(
        responses.POST,
        f"http://testserver/api/v1/element/{elt.id}/transcription/",
        status=418,
    )

    with pytest.raises(ErrorResponse):
        mock_elements_worker.create_transcription(
            element=elt,
            text="i am a line",
            confidence=0.42,
        )

    assert len(responses.calls) == len(BASE_API_CALLS) + 1
    assert [
        (call.request.method, call.request.url) for call in responses.calls
    ] == BASE_API_CALLS + [
        ("POST", f"http://testserver/api/v1/element/{elt.id}/transcription/")
    ]


def test_create_transcription(responses, mock_elements_worker):
    elt = Element({"id": "12341234-1234-1234-1234-123412341234"})
    responses.add(
        responses.POST,
        f"http://testserver/api/v1/element/{elt.id}/transcription/",
        status=200,
        json={
            "id": "56785678-5678-5678-5678-567856785678",
            "text": "i am a line",
            "confidence": 0.42,
            "worker_run_id": "56785678-5678-5678-5678-567856785678",
        },
    )

    mock_elements_worker.create_transcription(
        element=elt,
        text="i am a line",
        confidence=0.42,
    )

    assert len(responses.calls) == len(BASE_API_CALLS) + 1
    assert [
        (call.request.method, call.request.url) for call in responses.calls
    ] == BASE_API_CALLS + [
        ("POST", f"http://testserver/api/v1/element/{elt.id}/transcription/"),
    ]

    assert json.loads(responses.calls[-1].request.body) == {
        "text": "i am a line",
        "worker_run_id": "56785678-5678-5678-5678-567856785678",
        "confidence": 0.42,
        "orientation": "horizontal-lr",
    }


def test_create_transcription_with_cache(responses, mock_elements_worker_with_cache):
    elt = CachedElement.create(id="12341234-1234-1234-1234-123412341234", type="thing")

    responses.add(
        responses.POST,
        f"http://testserver/api/v1/element/{elt.id}/transcription/",
        status=200,
        json={
            "id": "56785678-5678-5678-5678-567856785678",
            "text": "i am a line",
            "confidence": 0.42,
            "orientation": "horizontal-lr",
            "worker_run_id": "56785678-5678-5678-5678-567856785678",
        },
    )

    mock_elements_worker_with_cache.create_transcription(
        element=elt,
        text="i am a line",
        confidence=0.42,
    )

    assert len(responses.calls) == len(BASE_API_CALLS) + 1
    assert [
        (call.request.method, call.request.url) for call in responses.calls
    ] == BASE_API_CALLS + [
        ("POST", f"http://testserver/api/v1/element/{elt.id}/transcription/"),
    ]

    assert json.loads(responses.calls[-1].request.body) == {
        "text": "i am a line",
        "worker_run_id": "56785678-5678-5678-5678-567856785678",
        "orientation": "horizontal-lr",
        "confidence": 0.42,
    }

    # Check that created transcription was properly stored in SQLite cache
    assert list(CachedTranscription.select()) == [
        CachedTranscription(
            id=UUID("56785678-5678-5678-5678-567856785678"),
            element_id=UUID(elt.id),
            text="i am a line",
            confidence=0.42,
            orientation=TextOrientation.HorizontalLeftToRight,
            worker_version_id=None,
            worker_run_id=UUID("56785678-5678-5678-5678-567856785678"),
        )
    ]


def test_create_transcription_orientation_with_cache(
    responses, mock_elements_worker_with_cache
):
    elt = CachedElement.create(id="12341234-1234-1234-1234-123412341234", type="thing")
    responses.add(
        responses.POST,
        f"http://testserver/api/v1/element/{elt.id}/transcription/",
        status=200,
        json={
            "id": "56785678-5678-5678-5678-567856785678",
            "text": "Animula vagula blandula",
            "confidence": 0.42,
            "orientation": "vertical-lr",
            "worker_run_id": "56785678-5678-5678-5678-567856785678",
        },
    )
    mock_elements_worker_with_cache.create_transcription(
        element=elt,
        text="Animula vagula blandula",
        orientation=TextOrientation.VerticalLeftToRight,
        confidence=0.42,
    )
    assert json.loads(responses.calls[-1].request.body) == {
        "text": "Animula vagula blandula",
        "worker_run_id": "56785678-5678-5678-5678-567856785678",
        "orientation": "vertical-lr",
        "confidence": 0.42,
    }
    # Check that the text orientation was properly stored in SQLite cache
    assert list(map(model_to_dict, CachedTranscription.select())) == [
        {
            "id": UUID("56785678-5678-5678-5678-567856785678"),
            "element": {
                "id": UUID("12341234-1234-1234-1234-123412341234"),
                "parent_id": None,
                "type": "thing",
                "image": None,
                "polygon": None,
                "rotation_angle": 0,
                "mirrored": False,
                "initial": False,
                "worker_version_id": None,
                "worker_run_id": None,
                "confidence": None,
            },
            "text": "Animula vagula blandula",
            "confidence": 0.42,
            "orientation": TextOrientation.VerticalLeftToRight.value,
            "worker_version_id": None,
            "worker_run_id": UUID("56785678-5678-5678-5678-567856785678"),
        }
    ]


def test_create_transcriptions_wrong_transcriptions(mock_elements_worker):
    with pytest.raises(
        AssertionError,
        match="transcriptions shouldn't be null and should be of type list",
    ):
        mock_elements_worker.create_transcriptions(
            transcriptions=None,
        )

    with pytest.raises(
        AssertionError,
        match="transcriptions shouldn't be null and should be of type list",
    ):
        mock_elements_worker.create_transcriptions(
            transcriptions=1234,
        )

    with pytest.raises(
        AssertionError,
        match="Transcription at index 1 in transcriptions: element_id shouldn't be null and should be of type str",
    ):
        mock_elements_worker.create_transcriptions(
            transcriptions=[
                {
                    "element_id": "11111111-1111-1111-1111-111111111111",
                    "text": "The",
                    "confidence": 0.75,
                },
                {
                    "text": "word",
                    "confidence": 0.5,
                },
            ],
        )

    with pytest.raises(
        AssertionError,
        match="Transcription at index 1 in transcriptions: element_id shouldn't be null and should be of type str",
    ):
        mock_elements_worker.create_transcriptions(
            transcriptions=[
                {
                    "element_id": "11111111-1111-1111-1111-111111111111",
                    "text": "The",
                    "confidence": 0.75,
                },
                {
                    "element_id": None,
                    "text": "word",
                    "confidence": 0.5,
                },
            ],
        )

    with pytest.raises(
        AssertionError,
        match="Transcription at index 1 in transcriptions: element_id shouldn't be null and should be of type str",
    ):
        mock_elements_worker.create_transcriptions(
            transcriptions=[
                {
                    "element_id": "11111111-1111-1111-1111-111111111111",
                    "text": "The",
                    "confidence": 0.75,
                },
                {
                    "element_id": 1234,
                    "text": "word",
                    "confidence": 0.5,
                },
            ],
        )

    with pytest.raises(
        AssertionError,
        match="Transcription at index 1 in transcriptions: text shouldn't be null and should be of type str",
    ):
        mock_elements_worker.create_transcriptions(
            transcriptions=[
                {
                    "element_id": "11111111-1111-1111-1111-111111111111",
                    "text": "The",
                    "confidence": 0.75,
                },
                {
                    "element_id": "11111111-1111-1111-1111-111111111111",
                    "confidence": 0.5,
                },
            ],
        )

    with pytest.raises(
        AssertionError,
        match="Transcription at index 1 in transcriptions: text shouldn't be null and should be of type str",
    ):
        mock_elements_worker.create_transcriptions(
            transcriptions=[
                {
                    "element_id": "11111111-1111-1111-1111-111111111111",
                    "text": "The",
                    "confidence": 0.75,
                },
                {
                    "element_id": "11111111-1111-1111-1111-111111111111",
                    "text": None,
                    "confidence": 0.5,
                },
            ],
        )

    with pytest.raises(
        AssertionError,
        match="Transcription at index 1 in transcriptions: text shouldn't be null and should be of type str",
    ):
        mock_elements_worker.create_transcriptions(
            transcriptions=[
                {
                    "element_id": "11111111-1111-1111-1111-111111111111",
                    "text": "The",
                    "confidence": 0.75,
                },
                {
                    "element_id": "11111111-1111-1111-1111-111111111111",
                    "text": 1234,
                    "confidence": 0.5,
                },
            ],
        )

    with pytest.raises(
        AssertionError,
        match=re.escape(
            "Transcription at index 1 in transcriptions: confidence shouldn't be null and should be a float in [0..1] range"
        ),
    ):
        mock_elements_worker.create_transcriptions(
            transcriptions=[
                {
                    "element_id": "11111111-1111-1111-1111-111111111111",
                    "text": "The",
                    "confidence": 0.75,
                },
                {
                    "element_id": "11111111-1111-1111-1111-111111111111",
                    "text": "word",
                },
            ],
        )

    with pytest.raises(
        AssertionError,
        match=re.escape(
            "Transcription at index 1 in transcriptions: confidence shouldn't be null and should be a float in [0..1] range"
        ),
    ):
        mock_elements_worker.create_transcriptions(
            transcriptions=[
                {
                    "element_id": "11111111-1111-1111-1111-111111111111",
                    "text": "The",
                    "confidence": 0.75,
                },
                {
                    "element_id": "11111111-1111-1111-1111-111111111111",
                    "text": "word",
                    "confidence": None,
                },
            ],
        )

    with pytest.raises(
        AssertionError,
        match=re.escape(
            "Transcription at index 1 in transcriptions: confidence shouldn't be null and should be a float in [0..1] range"
        ),
    ):
        mock_elements_worker.create_transcriptions(
            transcriptions=[
                {
                    "element_id": "11111111-1111-1111-1111-111111111111",
                    "text": "The",
                    "confidence": 0.75,
                },
                {
                    "element_id": "11111111-1111-1111-1111-111111111111",
                    "text": "word",
                    "confidence": "a wrong confidence",
                },
            ],
        )

    with pytest.raises(
        AssertionError,
        match=re.escape(
            "Transcription at index 1 in transcriptions: confidence shouldn't be null and should be a float in [0..1] range"
        ),
    ):
        mock_elements_worker.create_transcriptions(
            transcriptions=[
                {
                    "element_id": "11111111-1111-1111-1111-111111111111",
                    "text": "The",
                    "confidence": 0.75,
                },
                {
                    "element_id": "11111111-1111-1111-1111-111111111111",
                    "text": "word",
                    "confidence": 0,
                },
            ],
        )

    with pytest.raises(
        AssertionError,
        match=re.escape(
            "Transcription at index 1 in transcriptions: confidence shouldn't be null and should be a float in [0..1] range"
        ),
    ):
        mock_elements_worker.create_transcriptions(
            transcriptions=[
                {
                    "element_id": "11111111-1111-1111-1111-111111111111",
                    "text": "The",
                    "confidence": 0.75,
                },
                {
                    "element_id": "11111111-1111-1111-1111-111111111111",
                    "text": "word",
                    "confidence": 2.00,
                },
            ],
        )

    with pytest.raises(
        AssertionError,
        match="Transcription at index 1 in transcriptions: orientation shouldn't be null and should be of type TextOrientation",
    ):
        mock_elements_worker.create_transcriptions(
            transcriptions=[
                {
                    "element_id": "11111111-1111-1111-1111-111111111111",
                    "text": "The",
                    "confidence": 0.75,
                },
                {
                    "element_id": "11111111-1111-1111-1111-111111111111",
                    "text": "word",
                    "confidence": 0.28,
                    "orientation": "wobble",
                },
            ],
        )


def test_create_transcriptions_api_error(responses, mock_elements_worker):
    responses.add(
        responses.POST,
        "http://testserver/api/v1/transcription/bulk/",
        status=418,
    )
    trans = [
        {
            "element_id": "11111111-1111-1111-1111-111111111111",
            "text": "The",
            "confidence": 0.75,
        },
        {
            "element_id": "11111111-1111-1111-1111-111111111111",
            "text": "word",
            "confidence": 0.42,
        },
    ]

    with pytest.raises(ErrorResponse):
        mock_elements_worker.create_transcriptions(transcriptions=trans)

    assert len(responses.calls) == len(BASE_API_CALLS) + 1
    assert [
        (call.request.method, call.request.url) for call in responses.calls
    ] == BASE_API_CALLS + [("POST", "http://testserver/api/v1/transcription/bulk/")]


@pytest.mark.parametrize("batch_size", [DEFAULT_BATCH_SIZE, 1])
def test_create_transcriptions(batch_size, responses, mock_elements_worker_with_cache):
    CachedElement.create(id="11111111-1111-1111-1111-111111111111", type="thing")
    transcriptions = [
        {
            "element_id": "11111111-1111-1111-1111-111111111111",
            "text": "The",
            "confidence": 0.75,
        },
        {
            "element_id": "11111111-1111-1111-1111-111111111111",
            "text": "word",
            "confidence": 0.42,
        },
    ]

    if batch_size > 1:
        responses.add(
            responses.POST,
            "http://testserver/api/v1/transcription/bulk/",
            status=200,
            json={
                "worker_run_id": "56785678-5678-5678-5678-567856785678",
                "transcriptions": [
                    {
                        "id": "00000000-0000-0000-0000-000000000000",
                        "element_id": "11111111-1111-1111-1111-111111111111",
                        "text": "The",
                        "orientation": "horizontal-lr",
                        "confidence": 0.75,
                    },
                    {
                        "id": "11111111-1111-1111-1111-111111111111",
                        "element_id": "11111111-1111-1111-1111-111111111111",
                        "text": "word",
                        "orientation": "horizontal-lr",
                        "confidence": 0.42,
                    },
                ],
            },
        )
    else:
        for tr, tr_id in zip(
            transcriptions,
            [
                "00000000-0000-0000-0000-000000000000",
                "11111111-1111-1111-1111-111111111111",
            ],
            strict=False,
        ):
            responses.add(
                responses.POST,
                "http://testserver/api/v1/transcription/bulk/",
                status=200,
                json={
                    "worker_run_id": "56785678-5678-5678-5678-567856785678",
                    "transcriptions": [
                        {
                            "id": tr_id,
                            "element_id": tr["element_id"],
                            "text": tr["text"],
                            "orientation": "horizontal-lr",
                            "confidence": tr["confidence"],
                        }
                    ],
                },
            )

    mock_elements_worker_with_cache.create_transcriptions(
        transcriptions=transcriptions,
        batch_size=batch_size,
    )

    bulk_api_calls = [
        (
            "POST",
            "http://testserver/api/v1/transcription/bulk/",
        )
    ]
    if batch_size != DEFAULT_BATCH_SIZE:
        bulk_api_calls.append(
            (
                "POST",
                "http://testserver/api/v1/transcription/bulk/",
            )
        )

    assert len(responses.calls) == len(BASE_API_CALLS) + len(bulk_api_calls)
    assert [
        (call.request.method, call.request.url) for call in responses.calls
    ] == BASE_API_CALLS + bulk_api_calls

    first_tr = {
        **transcriptions[0],
        "orientation": TextOrientation.HorizontalLeftToRight.value,
    }
    second_tr = {
        **transcriptions[1],
        "orientation": TextOrientation.HorizontalLeftToRight.value,
    }
    empty_payload = {
        "transcriptions": [],
        "worker_run_id": "56785678-5678-5678-5678-567856785678",
    }

    bodies = []
    first_call_idx = None
    if batch_size > 1:
        first_call_idx = -1
        bodies.append({**empty_payload, "transcriptions": [first_tr, second_tr]})
    else:
        first_call_idx = -2
        bodies.append({**empty_payload, "transcriptions": [first_tr]})
        bodies.append({**empty_payload, "transcriptions": [second_tr]})

    assert [
        json.loads(bulk_call.request.body)
        for bulk_call in responses.calls[first_call_idx:]
    ] == bodies

    # Check that created transcriptions were properly stored in SQLite cache
    assert list(CachedTranscription.select()) == [
        CachedTranscription(
            id=UUID("00000000-0000-0000-0000-000000000000"),
            element_id=UUID("11111111-1111-1111-1111-111111111111"),
            text="The",
            confidence=0.75,
            orientation=TextOrientation.HorizontalLeftToRight,
            worker_run_id=UUID("56785678-5678-5678-5678-567856785678"),
        ),
        CachedTranscription(
            id=UUID("11111111-1111-1111-1111-111111111111"),
            element_id=UUID("11111111-1111-1111-1111-111111111111"),
            text="word",
            confidence=0.42,
            orientation=TextOrientation.HorizontalLeftToRight,
            worker_run_id=UUID("56785678-5678-5678-5678-567856785678"),
        ),
    ]


def test_create_transcriptions_orientation(responses, mock_elements_worker_with_cache):
    CachedElement.create(id="11111111-1111-1111-1111-111111111111", type="thing")
    trans = [
        {
            "element_id": "11111111-1111-1111-1111-111111111111",
            "text": "Animula vagula blandula",
            "confidence": 0.12,
            "orientation": TextOrientation.HorizontalRightToLeft,
        },
        {
            "element_id": "11111111-1111-1111-1111-111111111111",
            "text": "Hospes comesque corporis",
            "confidence": 0.21,
            "orientation": TextOrientation.VerticalLeftToRight,
        },
    ]

    responses.add(
        responses.POST,
        "http://testserver/api/v1/transcription/bulk/",
        status=200,
        json={
            "worker_run_id": "56785678-5678-5678-5678-567856785678",
            "transcriptions": [
                {
                    "id": "00000000-0000-0000-0000-000000000000",
                    "element_id": "11111111-1111-1111-1111-111111111111",
                    "text": "Animula vagula blandula",
                    "orientation": "horizontal-rl",
                    "confidence": 0.12,
                },
                {
                    "id": "11111111-1111-1111-1111-111111111111",
                    "element_id": "11111111-1111-1111-1111-111111111111",
                    "text": "Hospes comesque corporis",
                    "orientation": "vertical-lr",
                    "confidence": 0.21,
                },
            ],
        },
    )

    mock_elements_worker_with_cache.create_transcriptions(
        transcriptions=trans,
    )

    assert json.loads(responses.calls[-1].request.body) == {
        "worker_run_id": "56785678-5678-5678-5678-567856785678",
        "transcriptions": [
            {
                "element_id": "11111111-1111-1111-1111-111111111111",
                "text": "Animula vagula blandula",
                "confidence": 0.12,
                "orientation": TextOrientation.HorizontalRightToLeft.value,
            },
            {
                "element_id": "11111111-1111-1111-1111-111111111111",
                "text": "Hospes comesque corporis",
                "confidence": 0.21,
                "orientation": TextOrientation.VerticalLeftToRight.value,
            },
        ],
    }

    # Check that oriented transcriptions were properly stored in SQLite cache
    assert list(map(model_to_dict, CachedTranscription.select())) == [
        {
            "id": UUID("00000000-0000-0000-0000-000000000000"),
            "element": {
                "id": UUID("11111111-1111-1111-1111-111111111111"),
                "parent_id": None,
                "type": "thing",
                "image": None,
                "polygon": None,
                "rotation_angle": 0,
                "mirrored": False,
                "initial": False,
                "worker_version_id": None,
                "worker_run_id": None,
                "confidence": None,
            },
            "text": "Animula vagula blandula",
            "confidence": 0.12,
            "orientation": TextOrientation.HorizontalRightToLeft.value,
            "worker_version_id": None,
            "worker_run_id": UUID("56785678-5678-5678-5678-567856785678"),
        },
        {
            "id": UUID("11111111-1111-1111-1111-111111111111"),
            "element": {
                "id": UUID("11111111-1111-1111-1111-111111111111"),
                "parent_id": None,
                "type": "thing",
                "image": None,
                "polygon": None,
                "rotation_angle": 0,
                "mirrored": False,
                "initial": False,
                "worker_version_id": None,
                "worker_run_id": None,
                "confidence": None,
            },
            "text": "Hospes comesque corporis",
            "confidence": 0.21,
            "orientation": TextOrientation.VerticalLeftToRight.value,
            "worker_version_id": None,
            "worker_run_id": UUID("56785678-5678-5678-5678-567856785678"),
        },
    ]



import importlib.metadata
import logging

logging.basicConfig(
    level=logging.INFO,
    format="%(asctime)s %(levelname)s/%(name)s: %(message)s",
)
logger = logging.getLogger(__name__)

VERSION = importlib.metadata.version("arkindex-base-worker")



"""
Database mappings and helper methods for the experimental worker caching feature.

On methods that support caching, the database will be used for all reads,
and writes will go both to the Arkindex API and the database,
reducing network usage.
"""

import json
import sqlite3
from pathlib import Path

from peewee import (
    SQL,
    BooleanField,
    CharField,
    Check,
    CompositeKey,
    Field,
    FloatField,
    ForeignKeyField,
    IntegerField,
    Model,
    OperationalError,
    SqliteDatabase,
    TextField,
    UUIDField,
)
from PIL import Image

from arkindex_worker import logger

db = SqliteDatabase(None)


class JSONField(Field):
    """
    A Peewee field that stores a JSON payload as a string and parses it automatically.
    """

    field_type = "text"

    def db_value(self, value):
        if value is None:
            return
        return json.dumps(value)

    def python_value(self, value):
        if value is None:
            return
        return json.loads(value)


class Version(Model):
    """
    Cache version table, used to warn about incompatible cache databases
    when a worker uses an outdated version of ``base-worker``.
    """

    version = IntegerField(primary_key=True)

    class Meta:
        database = db
        table_name = "version"


class CachedImage(Model):
    """
    Cache image table
    """

    id = UUIDField(primary_key=True)
    width = IntegerField()
    height = IntegerField()
    url = TextField()
    version = IntegerField(default=2)

    class Meta:
        database = db
        table_name = "images"


class CachedElement(Model):
    """
    Cache element table
    """

    id = UUIDField(primary_key=True)
    parent_id = UUIDField(null=True)
    type = CharField(max_length=50)
    image = ForeignKeyField(CachedImage, backref="elements", null=True)
    polygon = JSONField(null=True)
    rotation_angle = IntegerField(default=0)
    mirrored = BooleanField(default=False)
    initial = BooleanField(default=False)
    # Needed to filter elements with cache
    worker_version_id = UUIDField(null=True)
    worker_run_id = UUIDField(null=True)
    confidence = FloatField(null=True)

    class Meta:
        database = db
        table_name = "elements"

    def open_image(
        self,
        *args,
        max_width: int | None = None,
        max_height: int | None = None,
        **kwargs,
    ) -> Image:
        """
        Open this element's image as a Pillow image.
        This does not crop the image to the element's polygon.
        IIIF servers with maxWidth, maxHeight or maxArea restrictions on image size are not supported.

        Warns:
        ----
           If both, ``max_width`` and ``max_height`` are set, the image ratio is not preserved.


        :param *args: Positional arguments passed to [arkindex_worker.image.open_image][]
        :param max_width: The maximum width of the image.
        :param max_height: The maximum height of the image.
        :param **kwargs: Keyword arguments passed to [arkindex_worker.image.open_image][]
        :raises ValueError: When this element does not have an image ID or a polygon.
        :return: A Pillow image.
        """
        from arkindex_worker.image import open_image, polygon_bounding_box

        if not self.image_id or not self.polygon:
            raise ValueError(f"Element {self.id} has no image")

        # Always fetch the image from the bounding box when size differs from full image
        bounding_box = polygon_bounding_box(self.polygon)
        if (
            bounding_box.width != self.image.width
            or bounding_box.height != self.image.height
        ):
            box = f"{bounding_box.x},{bounding_box.y},{bounding_box.width},{bounding_box.height}"
        else:
            box = "full"

        if max_width is None and max_height is None:
            resize = "full"
        else:
            if (
                # Do not resize for polygons that do not exactly match the images
                # as the resize is made directly by the IIIF server using the box parameter
                bounding_box.width != self.image.width
                or bounding_box.height != self.image.height
            ) or (
                # Do not resize when the image is below the maximum size
                (max_width is None or self.image.width <= max_width)
                and (max_height is None or self.image.height <= max_height)
            ):
                resize = "full"
            else:
                resize = f"{max_width or ''},{max_height or ''}"

        # Use `max` instead of `full` for IIIF 3, since `full` was deprecated in 2.1 then removed in 3.0
        if self.image.version == 3 and resize == "full":
            resize = "max"

        url = self.image.url
        if not url.endswith("/"):
            url += "/"

        return open_image(
            f"{url}{box}/{resize}/0/default.jpg",
            *args,
            rotation_angle=self.rotation_angle,
            mirrored=self.mirrored,
            **kwargs,
        )


class CachedTranscription(Model):
    """
    Cache transcription table
    """

    id = UUIDField(primary_key=True)
    element = ForeignKeyField(CachedElement, backref="transcriptions")
    text = TextField()
    confidence = FloatField(null=True)
    orientation = CharField(max_length=50)
    # Needed to filter transcriptions with cache
    worker_version_id = UUIDField(null=True)
    worker_run_id = UUIDField(null=True)

    class Meta:
        database = db
        table_name = "transcriptions"


class CachedClassification(Model):
    """
    Cache classification table
    """

    id = UUIDField(primary_key=True)
    element = ForeignKeyField(CachedElement, backref="classifications")
    class_name = TextField()
    confidence = FloatField()
    state = CharField(max_length=10)
    worker_run_id = UUIDField(null=True)

    class Meta:
        database = db
        table_name = "classifications"


class CachedTranscriptionEntity(Model):
    """
    Cache transcription entity table
    """

    transcription = ForeignKeyField(
        CachedTranscription, backref="transcription_entities"
    )
    type = CharField(max_length=50)
    offset = IntegerField(constraints=[Check("offset >= 0")])
    length = IntegerField(constraints=[Check("length > 0")])
    worker_run_id = UUIDField(null=True)
    confidence = FloatField(null=True)

    class Meta:
        primary_key = CompositeKey("transcription", "type")
        database = db
        table_name = "transcription_entities"


class CachedDataset(Model):
    id = UUIDField(primary_key=True)
    name = CharField()
    state = CharField(constraints=[SQL("DEFAULT 'open'")])
    sets = TextField()

    class Meta:
        database = db
        table_name = "datasets"


class CachedDatasetElement(Model):
    id = UUIDField(primary_key=True)
    element = ForeignKeyField(column_name="element_id", field="id", model=CachedElement)
    dataset = ForeignKeyField(column_name="dataset_id", field="id", model=CachedDataset)
    set_name = CharField()

    class Meta:
        database = db
        table_name = "dataset_elements"


# Add all the managed models in that list
# It's used here, but also in unit tests
MODELS = [
    CachedImage,
    CachedElement,
    CachedTranscription,
    CachedClassification,
    CachedTranscriptionEntity,
    CachedDataset,
    CachedDatasetElement,
]
SQL_VERSION = 5


def init_cache_db(path: Path):
    """
    Create the cache database on the given path
    :param path: Where the new database should be created
    """
    db.init(
        path,
        pragmas={
            # SQLite ignores foreign keys and check constraints by default!
            "foreign_keys": 1,
            "ignore_check_constraints": 0,
        },
    )
    db.connect()
    logger.info(f"Connected to cache on {path}")


def create_tables():
    """
    Creates the tables in the cache DB only if they do not already exist.
    """
    db.create_tables(MODELS)


def create_version_table():
    """
    Creates the Version table in the cache DB.
    This step must be independent from other tables creation since we only
    want to create the table and add the one and only Version entry when the
    cache is created from scratch.
    """
    db.create_tables([Version])
    Version.create(version=SQL_VERSION)


def check_version(cache_path: str | Path):
    """
    Check the validity of the SQLite version

    :param cache_path: Path towards a local SQLite database
    """
    with SqliteDatabase(cache_path) as provided_db, provided_db.bind_ctx([Version]):
        try:
            version = Version.get().version
        except OperationalError:
            version = None

        assert version == SQL_VERSION, (
            f"The SQLite database {cache_path} does not have the correct cache version, it should be {SQL_VERSION}"
        )


def merge_parents_cache(paths: list, current_database: Path):
    """
    Merge all the potential parent task's databases into the existing local one
    :param paths: Path to cache databases
    :param current_database: Path to the current database
    """
    assert current_database.exists()

    if not paths:
        logger.info("No parents cache to use")
        return

    # Open a connection on current database
    connection = sqlite3.connect(current_database)
    cursor = connection.cursor()

    # Merge each table into the local database
    for idx, path in enumerate(paths):
        # Check that the parent cache uses a compatible version
        check_version(path)

        with SqliteDatabase(path) as source, source.bind_ctx(MODELS):
            source.create_tables(MODELS)

        logger.info(f"Merging parent db {path} into {current_database}")
        statements = [
            "PRAGMA page_size=80000;",
            "PRAGMA synchronous=OFF;",
            f"ATTACH DATABASE '{path}' AS source_{idx};",
            f"REPLACE INTO images SELECT * FROM source_{idx}.images;",
            f"REPLACE INTO elements SELECT * FROM source_{idx}.elements;",
            f"REPLACE INTO transcriptions SELECT * FROM source_{idx}.transcriptions;",
            f"REPLACE INTO classifications SELECT * FROM source_{idx}.classifications;",
            f"REPLACE INTO transcription_entities SELECT * FROM source_{idx}.transcription_entities;",
            f"REPLACE INTO datasets SELECT * FROM source_{idx}.datasets;",
            f"REPLACE INTO dataset_elements SELECT * FROM source_{idx}.dataset_elements;",
        ]

        for statement in statements:
            cursor.execute(statement)
        connection.commit()


def unsupported_cache(func):
    def wrapper(self, *args, **kwargs):
        results = func(self, *args, **kwargs)

        if self.use_cache:
            logger.warning(
                f"This API helper `{func.__name__}` did not update the cache database"
            )

        return results

    return wrapper



"""
ElementsWorker methods for metadata.
"""

from enum import Enum

from arkindex_worker import logger
from arkindex_worker.cache import CachedElement, unsupported_cache
from arkindex_worker.models import Element
from arkindex_worker.utils import DEFAULT_BATCH_SIZE, batch_publication, make_batches


class MetaType(Enum):
    """
    Type of a metadata.
    """

    Text = "text"
    """
    A regular string with no special interpretation.
    """

    Markdown = "markdown"
    """
    A metadata with a string value that should be interpreted as Markdown content.
    HTML is allowed, but the allowed HTML tags are restricted for security reasons.
    """

    Date = "date"
    """
    A metadata with a string value that should be interpreted as a date.
    The date should be formatted as an ISO 8601 date (``YYYY-MM-DD``).
    """

    Location = "location"
    """
    A metadata with a string value that should be interpreted as a location.
    """

    Reference = "reference"
    """
    A metadata with a string value that should be interpreted as an external identifier
    to this element, for example to preserve a link to the original data before it was
    imported into Arkindex.
    """

    Numeric = "numeric"
    """
    A metadata with a floating point value.
    """

    URL = "url"
    """
    A metadata with a string value that should be interpreted as a URL.
    Only the ``http`` and ``https`` schemes are allowed.
    """


class MetaDataMixin:
    @unsupported_cache
    def create_metadata(
        self,
        element: Element | CachedElement,
        type: MetaType,
        name: str,
        value: str,
    ) -> str:
        """
        Create a metadata on the given element through API.

        :param element: The element to create a metadata on.
        :param type: Type of the metadata.
        :param name: Name of the metadata.
        :param value: Value of the metadata.
        :returns: UUID of the created metadata.
        """
        assert element and isinstance(element, Element | CachedElement), (
            "element shouldn't be null and should be of type Element or CachedElement"
        )
        assert type and isinstance(type, MetaType), (
            "type shouldn't be null and should be of type MetaType"
        )
        assert name and isinstance(name, str), (
            "name shouldn't be null and should be of type str"
        )
        assert value and isinstance(value, str), (
            "value shouldn't be null and should be of type str"
        )
        if self.is_read_only:
            logger.warning("Cannot create metadata as this worker is in read-only mode")
            return

        metadata = self.api_client.request(
            "CreateMetaData",
            id=element.id,
            body={
                "type": type.value,
                "name": name,
                "value": value,
                "worker_run_id": self.worker_run_id,
            },
        )

        return metadata["id"]

    @unsupported_cache
    @batch_publication
    def create_metadata_bulk(
        self,
        element: Element | CachedElement,
        metadata_list: list[dict[str, MetaType | str | int | float | None]],
        batch_size: int = DEFAULT_BATCH_SIZE,
    ) -> list[dict[str, str]]:
        """
        Create multiple metadata on an existing element.
        This method does not support cache.

        :param element: The element to create multiple metadata on.
        :param metadata_list: The list of dict whose keys are the following:
            - type: MetaType
            - name: str
            - value: str | int | float
        :param batch_size: The size of each batch, which will be used to split the publication to avoid API errors.

        :returns: A list of dicts as returned in the ``metadata_list`` field by the ``CreateMetaDataBulk`` API endpoint.
        """
        assert element and isinstance(element, Element | CachedElement), (
            "element shouldn't be null and should be of type Element or CachedElement"
        )

        assert metadata_list and isinstance(metadata_list, list), (
            "metadata_list shouldn't be null and should be of type list of dict"
        )

        # Make a copy to avoid modifying the metadata_list argument
        metas = []
        for index, metadata in enumerate(metadata_list):
            assert isinstance(metadata, dict), (
                f"Element at index {index} in metadata_list: Should be of type dict"
            )

            assert metadata.get("type") and isinstance(
                metadata.get("type"), MetaType
            ), "type shouldn't be null and should be of type MetaType"

            assert metadata.get("name") and isinstance(metadata.get("name"), str), (
                "name shouldn't be null and should be of type str"
            )

            assert metadata.get("value") is not None and isinstance(
                metadata.get("value"), str | float | int
            ), "value shouldn't be null and should be of type (str or float or int)"

            metas.append(
                {
                    "type": metadata.get("type").value,
                    "name": metadata.get("name"),
                    "value": metadata.get("value"),
                }
            )

        if self.is_read_only:
            logger.warning("Cannot create metadata as this worker is in read-only mode")
            return

        created_metadata_list = [
            created_metadata
            for batch in make_batches(metas, "metadata", batch_size)
            for created_metadata in self.api_client.request(
                "CreateMetaDataBulk",
                id=element.id,
                body={
                    "worker_run_id": self.worker_run_id,
                    "metadata_list": batch,
                },
            )["metadata_list"]
        ]

        return created_metadata_list

    def list_element_metadata(
        self, element: Element | CachedElement, load_parents: bool | None = None
    ) -> list[dict[str, str]]:
        """
        List all metadata linked to an element.
        This method does not support cache.

        :param element: The element to list metadata on.
        :param load_parents: Also include all metadata from the element's parents in the response.
        """
        assert element and isinstance(element, Element | CachedElement), (
            "element shouldn't be null and should be of type Element or CachedElement"
        )

        query_params = {}
        if load_parents is not None:
            assert isinstance(load_parents, bool), "load_parents should be of type bool"
            query_params["load_parents"] = load_parents

        return self.api_client.paginate(
            "ListElementMetaData", id=element.id, **query_params
        )



import json
import re

import pytest

from arkindex.exceptions import ErrorResponse
from arkindex_worker.cache import CachedElement
from arkindex_worker.models import Element
from arkindex_worker.utils import DEFAULT_BATCH_SIZE
from arkindex_worker.worker import MetaType

from . import BASE_API_CALLS


def test_create_metadata_wrong_element(mock_elements_worker):
    with pytest.raises(
        AssertionError,
        match="element shouldn't be null and should be of type Element or CachedElement",
    ):
        mock_elements_worker.create_metadata(
            element=None,
            type=MetaType.Location,
            name="Teklia",
            value="La Turbine, Grenoble 38000",
        )

    with pytest.raises(
        AssertionError,
        match="element shouldn't be null and should be of type Element or CachedElement",
    ):
        mock_elements_worker.create_metadata(
            element="not element type",
            type=MetaType.Location,
            name="Teklia",
            value="La Turbine, Grenoble 38000",
        )


def test_create_metadata_wrong_type(mock_elements_worker):
    elt = Element({"id": "12341234-1234-1234-1234-123412341234"})

    with pytest.raises(
        AssertionError, match="type shouldn't be null and should be of type MetaType"
    ):
        mock_elements_worker.create_metadata(
            element=elt,
            type=None,
            name="Teklia",
            value="La Turbine, Grenoble 38000",
        )

    with pytest.raises(
        AssertionError, match="type shouldn't be null and should be of type MetaType"
    ):
        mock_elements_worker.create_metadata(
            element=elt,
            type=1234,
            name="Teklia",
            value="La Turbine, Grenoble 38000",
        )

    with pytest.raises(
        AssertionError, match="type shouldn't be null and should be of type MetaType"
    ):
        mock_elements_worker.create_metadata(
            element=elt,
            type="not_a_metadata_type",
            name="Teklia",
            value="La Turbine, Grenoble 38000",
        )


def test_create_metadata_wrong_name(mock_elements_worker):
    elt = Element({"id": "12341234-1234-1234-1234-123412341234"})

    with pytest.raises(
        AssertionError, match="name shouldn't be null and should be of type str"
    ):
        mock_elements_worker.create_metadata(
            element=elt,
            type=MetaType.Location,
            name=None,
            value="La Turbine, Grenoble 38000",
        )

    with pytest.raises(
        AssertionError, match="name shouldn't be null and should be of type str"
    ):
        mock_elements_worker.create_metadata(
            element=elt,
            type=MetaType.Location,
            name=1234,
            value="La Turbine, Grenoble 38000",
        )


def test_create_metadata_wrong_value(mock_elements_worker):
    elt = Element({"id": "12341234-1234-1234-1234-123412341234"})

    with pytest.raises(
        AssertionError, match="value shouldn't be null and should be of type str"
    ):
        mock_elements_worker.create_metadata(
            element=elt,
            type=MetaType.Location,
            name="Teklia",
            value=None,
        )

    with pytest.raises(
        AssertionError, match="value shouldn't be null and should be of type str"
    ):
        mock_elements_worker.create_metadata(
            element=elt,
            type=MetaType.Location,
            name="Teklia",
            value=1234,
        )


def test_create_metadata_api_error(responses, mock_elements_worker):
    elt = Element({"id": "12341234-1234-1234-1234-123412341234"})
    responses.add(
        responses.POST,
        "http://testserver/api/v1/element/12341234-1234-1234-1234-123412341234/metadata/",
        status=418,
    )

    with pytest.raises(ErrorResponse):
        mock_elements_worker.create_metadata(
            element=elt,
            type=MetaType.Location,
            name="Teklia",
            value="La Turbine, Grenoble 38000",
        )

    assert len(responses.calls) == len(BASE_API_CALLS) + 1
    assert [
        (call.request.method, call.request.url) for call in responses.calls
    ] == BASE_API_CALLS + [
        (
            "POST",
            "http://testserver/api/v1/element/12341234-1234-1234-1234-123412341234/metadata/",
        )
    ]


def test_create_metadata(responses, mock_elements_worker):
    elt = Element({"id": "12341234-1234-1234-1234-123412341234"})
    responses.add(
        responses.POST,
        "http://testserver/api/v1/element/12341234-1234-1234-1234-123412341234/metadata/",
        status=200,
        json={"id": "12345678-1234-1234-1234-123456789123"},
    )

    metadata_id = mock_elements_worker.create_metadata(
        element=elt,
        type=MetaType.Location,
        name="Teklia",
        value="La Turbine, Grenoble 38000",
    )

    assert len(responses.calls) == len(BASE_API_CALLS) + 1
    assert [
        (call.request.method, call.request.url) for call in responses.calls
    ] == BASE_API_CALLS + [
        (
            "POST",
            "http://testserver/api/v1/element/12341234-1234-1234-1234-123412341234/metadata/",
        ),
    ]
    assert json.loads(responses.calls[-1].request.body) == {
        "type": "location",
        "name": "Teklia",
        "value": "La Turbine, Grenoble 38000",
        "worker_run_id": "56785678-5678-5678-5678-567856785678",
    }
    assert metadata_id == "12345678-1234-1234-1234-123456789123"


def test_create_metadata_cached_element(responses, mock_elements_worker_with_cache):
    elt = CachedElement.create(id="12341234-1234-1234-1234-123412341234", type="thing")
    responses.add(
        responses.POST,
        "http://testserver/api/v1/element/12341234-1234-1234-1234-123412341234/metadata/",
        status=200,
        json={"id": "12345678-1234-1234-1234-123456789123"},
    )

    metadata_id = mock_elements_worker_with_cache.create_metadata(
        element=elt,
        type=MetaType.Location,
        name="Teklia",
        value="La Turbine, Grenoble 38000",
    )

    assert len(responses.calls) == len(BASE_API_CALLS) + 1
    assert [
        (call.request.method, call.request.url) for call in responses.calls
    ] == BASE_API_CALLS + [
        (
            "POST",
            "http://testserver/api/v1/element/12341234-1234-1234-1234-123412341234/metadata/",
        ),
    ]
    assert json.loads(responses.calls[-1].request.body) == {
        "type": "location",
        "name": "Teklia",
        "value": "La Turbine, Grenoble 38000",
        "worker_run_id": "56785678-5678-5678-5678-567856785678",
    }
    assert metadata_id == "12345678-1234-1234-1234-123456789123"


@pytest.mark.parametrize("batch_size", [DEFAULT_BATCH_SIZE, 1])
def test_create_metadata_bulk(batch_size, responses, mock_elements_worker):
    element = Element({"id": "12341234-1234-1234-1234-123412341234"})

    metadata_list = [
        {"type": MetaType.Text, "name": "fake_name", "value": "fake_value"},
        {
            "type": MetaType.Text,
            "name": "Year",
            "value": "2024",
        },
    ]
    if batch_size > 1:
        responses.add(
            responses.POST,
            "http://testserver/api/v1/element/12341234-1234-1234-1234-123412341234/metadata/bulk/",
            status=201,
            json={
                "worker_run_id": "56785678-5678-5678-5678-567856785678",
                "metadata_list": [
                    {
                        "id": "fake_metadata_id1",
                        "type": metadata_list[0]["type"].value,
                        "name": metadata_list[0]["name"],
                        "value": metadata_list[0]["value"],
                        "dates": [],
                    },
                    {
                        "id": "fake_metadata_id2",
                        "type": metadata_list[1]["type"].value,
                        "name": metadata_list[1]["name"],
                        "value": metadata_list[1]["value"],
                        "dates": [],
                    },
                ],
            },
        )
    else:
        for idx, meta in enumerate(metadata_list, start=1):
            responses.add(
                responses.POST,
                "http://testserver/api/v1/element/12341234-1234-1234-1234-123412341234/metadata/bulk/",
                status=201,
                json={
                    "worker_run_id": "56785678-5678-5678-5678-567856785678",
                    "metadata_list": [
                        {
                            "id": f"fake_metadata_id{idx}",
                            "type": meta["type"].value,
                            "name": meta["name"],
                            "value": meta["value"],
                            "dates": [],
                        }
                    ],
                },
            )

    created_metadata_list = mock_elements_worker.create_metadata_bulk(
        element, metadata_list, batch_size=batch_size
    )

    bulk_api_calls = [
        (
            "POST",
            "http://testserver/api/v1/element/12341234-1234-1234-1234-123412341234/metadata/bulk/",
        )
    ]
    if batch_size != DEFAULT_BATCH_SIZE:
        bulk_api_calls.append(
            (
                "POST",
                "http://testserver/api/v1/element/12341234-1234-1234-1234-123412341234/metadata/bulk/",
            )
        )

    assert len(responses.calls) == len(BASE_API_CALLS) + len(bulk_api_calls)
    assert [
        (call.request.method, call.request.url) for call in responses.calls
    ] == BASE_API_CALLS + bulk_api_calls

    first_meta = {
        **metadata_list[0],
        "type": metadata_list[0]["type"].value,
    }
    second_meta = {**metadata_list[1], "type": metadata_list[1]["type"].value}
    empty_payload = {
        "worker_run_id": "56785678-5678-5678-5678-567856785678",
        "metadata_list": [],
    }

    bodies = []
    first_call_idx = None
    if batch_size > 1:
        first_call_idx = -1
        bodies.append({**empty_payload, "metadata_list": [first_meta, second_meta]})
    else:
        first_call_idx = -2
        bodies.append({**empty_payload, "metadata_list": [first_meta]})
        bodies.append({**empty_payload, "metadata_list": [second_meta]})

    assert [
        json.loads(bulk_call.request.body)
        for bulk_call in responses.calls[first_call_idx:]
    ] == bodies

    assert created_metadata_list == [
        {
            "id": "fake_metadata_id1",
            "type": metadata_list[0]["type"].value,
            "name": metadata_list[0]["name"],
            "value": metadata_list[0]["value"],
            "dates": [],
        },
        {
            "id": "fake_metadata_id2",
            "type": metadata_list[1]["type"].value,
            "name": metadata_list[1]["name"],
            "value": metadata_list[1]["value"],
            "dates": [],
        },
    ]


@pytest.mark.parametrize("batch_size", [DEFAULT_BATCH_SIZE, 1])
def test_create_metadata_bulk_cached_element(
    batch_size, responses, mock_elements_worker_with_cache
):
    element = CachedElement.create(
        id="12341234-1234-1234-1234-123412341234", type="thing"
    )

    metadata_list = [
        {"type": MetaType.Text, "name": "fake_name", "value": "fake_value"},
        {
            "type": MetaType.Text,
            "name": "Year",
            "value": "2024",
        },
    ]
    if batch_size > 1:
        responses.add(
            responses.POST,
            "http://testserver/api/v1/element/12341234-1234-1234-1234-123412341234/metadata/bulk/",
            status=201,
            json={
                "worker_run_id": "56785678-5678-5678-5678-567856785678",
                "metadata_list": [
                    {
                        "id": "fake_metadata_id1",
                        "type": metadata_list[0]["type"].value,
                        "name": metadata_list[0]["name"],
                        "value": metadata_list[0]["value"],
                        "dates": [],
                    },
                    {
                        "id": "fake_metadata_id2",
                        "type": metadata_list[1]["type"].value,
                        "name": metadata_list[1]["name"],
                        "value": metadata_list[1]["value"],
                        "dates": [],
                    },
                ],
            },
        )
    else:
        for idx, meta in enumerate(metadata_list, start=1):
            responses.add(
                responses.POST,
                "http://testserver/api/v1/element/12341234-1234-1234-1234-123412341234/metadata/bulk/",
                status=201,
                json={
                    "worker_run_id": "56785678-5678-5678-5678-567856785678",
                    "metadata_list": [
                        {
                            "id": f"fake_metadata_id{idx}",
                            "type": meta["type"].value,
                            "name": meta["name"],
                            "value": meta["value"],
                            "dates": [],
                        }
                    ],
                },
            )

    created_metadata_list = mock_elements_worker_with_cache.create_metadata_bulk(
        element, metadata_list, batch_size=batch_size
    )

    bulk_api_calls = [
        (
            "POST",
            "http://testserver/api/v1/element/12341234-1234-1234-1234-123412341234/metadata/bulk/",
        )
    ]
    if batch_size != DEFAULT_BATCH_SIZE:
        bulk_api_calls.append(
            (
                "POST",
                "http://testserver/api/v1/element/12341234-1234-1234-1234-123412341234/metadata/bulk/",
            )
        )

    assert len(responses.calls) == len(BASE_API_CALLS) + len(bulk_api_calls)
    assert [
        (call.request.method, call.request.url) for call in responses.calls
    ] == BASE_API_CALLS + bulk_api_calls

    first_meta = {
        **metadata_list[0],
        "type": metadata_list[0]["type"].value,
    }
    second_meta = {**metadata_list[1], "type": metadata_list[1]["type"].value}
    empty_payload = {
        "worker_run_id": "56785678-5678-5678-5678-567856785678",
        "metadata_list": [],
    }

    bodies = []
    first_call_idx = None
    if batch_size > 1:
        first_call_idx = -1
        bodies.append({**empty_payload, "metadata_list": [first_meta, second_meta]})
    else:
        first_call_idx = -2
        bodies.append({**empty_payload, "metadata_list": [first_meta]})
        bodies.append({**empty_payload, "metadata_list": [second_meta]})

    assert [
        json.loads(bulk_call.request.body)
        for bulk_call in responses.calls[first_call_idx:]
    ] == bodies

    assert created_metadata_list == [
        {
            "id": "fake_metadata_id1",
            "type": metadata_list[0]["type"].value,
            "name": metadata_list[0]["name"],
            "value": metadata_list[0]["value"],
            "dates": [],
        },
        {
            "id": "fake_metadata_id2",
            "type": metadata_list[1]["type"].value,
            "name": metadata_list[1]["name"],
            "value": metadata_list[1]["value"],
            "dates": [],
        },
    ]


@pytest.mark.parametrize("wrong_element", [None, "not_element_type", 1234, 12.5])
def test_create_metadata_bulk_wrong_element(mock_elements_worker, wrong_element):
    wrong_metadata_list = [
        {"type": MetaType.Text, "name": "fake_name", "value": "fake_value"}
    ]
    with pytest.raises(
        AssertionError,
        match="element shouldn't be null and should be of type Element or CachedElement",
    ):
        mock_elements_worker.create_metadata_bulk(
            element=wrong_element, metadata_list=wrong_metadata_list
        )


@pytest.mark.parametrize("wrong_type", [None, "not_metadata_type", 1234, 12.5])
def test_create_metadata_bulk_wrong_type(mock_elements_worker, wrong_type):
    element = Element({"id": "12341234-1234-1234-1234-123412341234"})
    wrong_metadata_list = [
        {"type": wrong_type, "name": "fake_name", "value": "fake_value"}
    ]
    with pytest.raises(
        AssertionError, match="type shouldn't be null and should be of type MetaType"
    ):
        mock_elements_worker.create_metadata_bulk(
            element=element, metadata_list=wrong_metadata_list
        )


@pytest.mark.parametrize("wrong_name", [None, 1234, 12.5, [1, 2, 3, 4]])
def test_create_metadata_bulk_wrong_name(mock_elements_worker, wrong_name):
    element = Element({"id": "fake_element_id"})
    wrong_metadata_list = [
        {"type": MetaType.Text, "name": wrong_name, "value": "fake_value"}
    ]
    with pytest.raises(
        AssertionError, match="name shouldn't be null and should be of type str"
    ):
        mock_elements_worker.create_metadata_bulk(
            element=element, metadata_list=wrong_metadata_list
        )


@pytest.mark.parametrize("wrong_value", [None, [1, 2, 3, 4]])
def test_create_metadata_bulk_wrong_value(mock_elements_worker, wrong_value):
    element = Element({"id": "fake_element_id"})
    wrong_metadata_list = [
        {"type": MetaType.Text, "name": "fake_name", "value": wrong_value}
    ]
    with pytest.raises(
        AssertionError,
        match=re.escape(
            "value shouldn't be null and should be of type (str or float or int)"
        ),
    ):
        mock_elements_worker.create_metadata_bulk(
            element=element, metadata_list=wrong_metadata_list
        )


def test_create_metadata_bulk_api_error(responses, mock_elements_worker):
    element = Element({"id": "12341234-1234-1234-1234-123412341234"})
    metadata_list = [
        {
            "type": MetaType.Text,
            "name": "fake_name",
            "value": "fake_value",
        }
    ]
    responses.add(
        responses.POST,
        "http://testserver/api/v1/element/12341234-1234-1234-1234-123412341234/metadata/bulk/",
        status=418,
    )

    with pytest.raises(ErrorResponse):
        mock_elements_worker.create_metadata_bulk(element, metadata_list)

    assert len(responses.calls) == len(BASE_API_CALLS) + 1
    assert [
        (call.request.method, call.request.url) for call in responses.calls
    ] == BASE_API_CALLS + [
        (
            "POST",
            "http://testserver/api/v1/element/12341234-1234-1234-1234-123412341234/metadata/bulk/",
        )
    ]


def test_list_element_metadata_wrong_load_parents(fake_dummy_worker):
    element = Element({"id": "element_id"})
    with pytest.raises(AssertionError, match="load_parents should be of type bool"):
        fake_dummy_worker.list_element_metadata(
            element=element,
            load_parents="not bool",
        )


def test_list_element_metadata(fake_dummy_worker):
    element = Element({"id": "element_id"})
    fake_dummy_worker.api_client.add_response(
        "ListElementMetaData",
        id=element.id,
        response=[{"id": "metadata_id"}],
    )
    assert fake_dummy_worker.list_element_metadata(element) == [{"id": "metadata_id"}]

    assert len(fake_dummy_worker.api_client.history) == 1
    assert len(fake_dummy_worker.api_client.responses) == 0


def test_list_element_metadata_cached_element(
    responses, mock_elements_worker_with_cache
):
    element = CachedElement.create(id="element_id", type="thing")

    responses.add(
        responses.GET,
        "http://testserver/api/v1/element/element_id/metadata/",
        status=200,
        json=[{"id": "metadata_id"}],
    )

    assert mock_elements_worker_with_cache.list_element_metadata(element) == [
        {"id": "metadata_id"}
    ]

    assert len(responses.calls) == len(BASE_API_CALLS) + 1
    assert [
        (call.request.method, call.request.url) for call in responses.calls
    ] == BASE_API_CALLS + [
        (
            "GET",
            "http://testserver/api/v1/element/element_id/metadata/",
        )
    ]


def test_list_element_metadata_with_load_parents(fake_dummy_worker):
    element = Element({"id": "element_id"})
    fake_dummy_worker.api_client.add_response(
        "ListElementMetaData",
        id=element.id,
        load_parents=True,
        response=[{"id": "metadata_id"}, {"id": "parent_metadata_id"}],
    )
    assert fake_dummy_worker.list_element_metadata(element, load_parents=True) == [
        {"id": "metadata_id"},
        {"id": "parent_metadata_id"},
    ]

    assert len(fake_dummy_worker.api_client.history) == 1
    assert len(fake_dummy_worker.api_client.responses) == 0
